(******************************************************************

   Separation logic toolkit

   Andrew McCreight
   HASP Project
   Portland State University

   http://hasp.cs.pdx.edu/

   Copyright 2009 Portland State University

   This file is part of the separation logic toolkit.

   The separation logic toolkit is free software: you can redistribute
   it and/or modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation, either
   version 3 of the License, or (at your option) any later version.

   The separation logic toolkit is distributed in the hope that it
   will be useful, but WITHOUT ANY WARRANTY; without even the implied
   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with the separation logic toolkit.  If not, see
   <http://www.gnu.org/licenses/>.

*******************************************************************)


Require Export Word.
Require Import Memory.

Set Implicit Arguments.



Open Scope word_scope.
Delimit Scope addr_scope with addr.
Open Scope addr_scope.


Definition addrAdd (a:addr) (k':word) : addr := 
  let (b, k) := a in
    (b, k + k')%word.
Infix "+" := addrAdd : addr_scope.


Lemma addrAddAssoc : forall x a b,
  x + a + b = x + (a + b)%word.
Proof.
  destruct x as [p c]; intros a b; simpl; f_equal; ring.
Qed.


Lemma addrInj : forall x a b,
  a = b ->
  x + a = x + b.
Proof.
  intros x a b E; rewrite E; reflexivity.
Qed.


Definition addrSub (a:addr) (k':word) : addr := 
  let (b, k) := a in
    (b, k - k')%word.

Infix "-" := addrSub : addr_scope.


Lemma addrSubAssoc : forall x a b,
  x + a - b = x + (a - b)%word.
Proof.
  destruct x as [p c]; intros a b; simpl; f_equal; ring.
Qed.


Definition addrLe (a a':addr) : Prop :=
  let (b, k) := a in
  let (b', k') := a' in
    b = b' /\ (k <= k')%word.
Infix "<=" := addrLe : addr_scope.


Definition addrLt (a a':addr) :=
  let (b, k) := a in
  let (b', k') := a' in
    b = b' /\ (k < k')%word.
Infix "<" := addrLt : addr_scope.


(* for some reason injection doesn't always produce nice results *)
Lemma addrInjHelper : forall (b:block) (k k':word),
  (b, k) = (b, k') ->
  k = k'.
Proof.
  intros b k k' E; injector E; auto.
Qed.


Lemma plusFourZeroEq : forall x,
  x + 4 * 0 = x.
Proof.
  intros; destruct x; simpl; f_equal; ring.
Qed.
  
Hint Rewrite plusFourZeroEq : AddrSimpl.


Lemma addrPlusMinusEq : forall x y,
  x + y - y = x.
Proof.
  intros; destruct x; simpl; f_equal; ring.
Qed.

Lemma addrMinusPlusEq : forall x y,
  x - y + y = x.
Proof.
  intros; destruct x; simpl; f_equal; ring.
Qed.
Hint Rewrite addrPlusMinusEq addrMinusPlusEq : AddrSimpl.


Ltac asimpl := autorewrite with AddrSimpl.


Lemma addrAddInj : forall x y z,
  x + y = x + z ->
  y = z.
Proof.
  clear.  intros [b x] y z E; simpl in E.
  apply addrInjHelper in E.
  apply wordAddInj in E; auto.
Qed.
  

Lemma multAddEqZero : forall y,
  (y < windexBound)%word -> forall x,
    x = x + 4 * y ->
    y = 0.
Proof.
  intros y LT x E.
  apply wordIndexMultInj; auto.
  compute; reflexivity.
  pattern x at 1 in E; rewrite <- (plusFourZeroEq x) in E.
  apply addrAddInj in E; auto.
Qed.



(* reasoning about the block of addresses is useful when
   reasoning about address operations *)

(* might want to redefine addr as a record *)
Definition blockOf (x:addr) : block := fst x.

Lemma blockOfAddEq : forall (x:addr) (y:word),
  blockOf (x + y) = blockOf x.
Proof.
  intros [b x] y; reflexivity.
Qed.

Lemma blockOfSubEq : forall (x:addr) (y:word),
  blockOf (x - y) = blockOf x.
Proof.
  intros [b x] y; reflexivity.
Qed.

Hint Rewrite blockOfAddEq blockOfSubEq : BlockOf.
  

Definition offsetOf (x:addr) := snd x.


Definition blockAddr (b:block) : addr := (b, 0).
Coercion blockAddr : CM.block >-> addr.

Hint Rewrite natWordAddEq  : AddrSimpl.
Hint Rewrite <- addrAddAssoc : AddrSimpl.


(* ************************ *)

(* various simplifications that should probably be moved to Word *)

(* fail if k is not a constant *)
Ltac isNotConst k :=
  lazymatch k with
    | 0%nat => fail "term is nat constant"
    | S ?E => isNotConst E
    | _ => idtac
  end.

(* convert "S e" to "(e + 1)%word", but only when e is not a constant *)
Tactic Notation "unSucc" :=
  match goal with
    | |- context [S ?E] =>
      isNotConst E; rewrite (nat_to_wordSEq E)
  end.

Hint Rewrite word_to_nat_to_word : WordSimpl.

Tactic Notation "wordSimpl" :=
  repeat (progress unSucc);
  autorewrite with WordSimpl.

Tactic Notation "unSucc" "in" hyp(H) :=
  match type of H with
    | context [S ?E] =>
      isNotConst E; rewrite (nat_to_wordSEq E) in H
  end.

Tactic Notation "wordSimpl" "in" hyp(H) :=
  repeat (progress (unSucc in H));
  autorewrite with WordSimpl in H.


(* ************************ *)

(* try to turn things into word math, then use ring *)
Hint Rewrite natWordSubEq using omega : WordEqSolve.
Ltac wordEqSolve :=
  autorewrite with WordEqSolve; 
  repeat (progress unSucc);
  try ring.


(* tactic for solving equality of addresses (with equal base addrs?) *)
Lemma VptrEqI : forall a a',
  a = a' ->
  Vptr a = Vptr a'.
Proof.
  intros; subst; reflexivity.
Qed.
    
Hint Rewrite addrAddAssoc addrSubAssoc : AddrEqSolve.
(*Ltac addrEqSolve :=
  autorewrite with AddrEqSolve; try (apply VptrEqI);
    try (apply addrInj); try wordEqSolve; try reflexivity.*)

Lemma addrMinusNegEq : forall x a,
  x - a = x + (- a).
Proof.
  intros [b k] b'. simpl. f_equal. ring.
Qed.
Hint Rewrite addrMinusNegEq : AddrEqSolve.

Lemma plusZeroEq : forall a,
  a + 0 = a.
Proof.
  intros [b k]; simpl; f_equal; ring.
Qed.
Hint Rewrite plusZeroEq : AddrEqSolve.

Lemma plusZeroEq' : forall k,
  k = 0 -> forall a,
  a + k = a.
Proof.
  intros; subst; apply plusZeroEq.
Qed.


Ltac addrEqSolve :=
  autorewrite with AddrEqSolve;
    try (apply VptrEqI);
    match goal with
      | |- ?a = ?a => reflexivity
        (* we do this via pattern matching instead of just trying
           reflexivity because it can be very slow for some reason *)
      | |- ?a + _ = ?a + _ => apply addrInj; try wordEqSolve
      | |- ?a = ?a + _ => symmetry; apply plusZeroEq'; try wordEqSolve
      | |- ?a + _ = ?a => apply plusZeroEq'; try wordEqSolve
      | |- _ => idtac
    end.


