(******************************************************************

   Separation logic toolkit

   Andrew McCreight
   HASP Project
   Portland State University

   http://hasp.cs.pdx.edu/

   Copyright 2009 Portland State University

   This file is part of the separation logic toolkit.

   The separation logic toolkit is free software: you can redistribute
   it and/or modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation, either
   version 3 of the License, or (at your option) any later version.

   The separation logic toolkit is distributed in the hope that it
   will be useful, but WITHOUT ANY WARRANTY; without even the implied
   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with the separation logic toolkit.  If not, see
   <http://www.gnu.org/licenses/>.

*******************************************************************)


Require Import CoreLog.
Require Import ListNatSet.
Require Import CoreMemorySig.
Require Import FSets.
Require Import FnMap.
Require Import FnPMap.
Require Import ZArith.


Set Implicit Arguments.


(* based on Leroy-Blazy's paper on a memory model
http://pauillac.inria.fr/~xleroy/publi/memory-model-journal.pdf

Extended to better support separation reasoning.
  - each block had a set of valid subaddresses, instead of a range
  - defines a union of memories
  - defines unary memory
  - intrinsically canonical

*)



Module CM : CoreMemorySig with Definition block := nat.

  Definition block := nat.

  Module BlockOrdered := Nat_as_OT.

  Definition addr := prod block word.
  Definition blockOf (a:addr) : block :=
    let (b, _) := a in b.
  Definition offsetOf (a:addr) : word :=
    let (_, i) := a in i.

  Inductive val : Set :=
    | Vundef : val
    | Vword : word -> val
    | Vptr : addr -> val.


  Inductive mem0 : Type := 
    mem0i : NS.t -> map block WS.t -> map block (pmap word val) -> mem0.
    (* we need the set of blocks to make validity decidable *)

  (* mem0 projection operations *)
  Definition validOf (m:mem0) := let (x, _, _) := m in x.
  Definition addrsOf (m:mem0) := let (_, x, _) := m in x.
  Definition valsOf (m:mem0) := let (_, _, x) := m in x.

  Inductive memOk (m:mem0) : Prop :=
    | memOki :
      (forall b, NS.In b (validOf m) -> map_get (addrsOf m) b <> WS.empty) ->
      (forall b, ~NS.In b (validOf m) -> map_get (addrsOf m) b = WS.empty) ->
      (forall b i, ~WS.In i (map_get (addrsOf m) b) -> 
        pmap_get (map_get (valsOf m) b) i = None) ->
      memOk m.

  Record mem' : Type := memi {mval:mem0; mok:memOk mval}.
  Definition mem := mem'.


  (* predicates *)
  Definition validBlock0 (m:mem0) (b:block) : Prop :=
    NS.In b (validOf m).
  
  Definition validBlock (m:mem) (b:block) : Prop :=
    validBlock0 (mval m) b.
    
  Definition blockOffs0 m b := map_get (addrsOf m) b.

  Definition blockOffs m b := blockOffs0 (mval m) b.

  Definition validPtr0 (m:mem0) (p:addr) : Prop :=
    validBlock0 m (blockOf p) /\ 
    WS.In (offsetOf p) (blockOffs0 m (blockOf p)).

  Definition validPtr (m:mem) (p:addr) : Prop :=
    validBlock m (blockOf p) /\ WS.In (offsetOf p) (blockOffs m (blockOf p)).

  Lemma validPtrEqv : forall m p,
    validPtr m p <-> validPtr0 (mval m) p.
  Proof.
    intros [m M] p; unfold validPtr, validBlock, blockOffs; simpl.
    unfold validPtr0; tauto.
  Qed.

  Lemma validPtrEqv1 : forall m p,
    validPtr m p -> validPtr0 (mval m) p.
  Proof.
    intros; destruct (validPtrEqv m p); auto.
  Qed.

  Lemma validPtrEqv2 : forall m p,
    validPtr0 (mval m) p -> validPtr m p.
  Proof.
    intros; destruct (validPtrEqv m p); auto.
  Qed.


  (* decidability of predicates *)
  Definition eq_block_dec := eq_nat_dec.


  Lemma validBlock0_dec : forall m b,
    {validBlock0 m b} + {~validBlock0 m b}.
  Proof.
    intros [v len vv] b; unfold validBlock0; simpl.
    apply NSE.Miff.In_dec.
  Qed.

  Lemma validBlock_dec : forall m b,
    {validBlock m b} + {~validBlock m b}.
  Proof.
    unfold validBlock; intros [m M] b; simpl; apply validBlock0_dec.
  Qed.


  Lemma validPtr0_dec : forall m p,
    {validPtr0 m p} + {~validPtr0 m p}.
  Proof.
    unfold validPtr0; intros m p.
    destruct (validBlock0_dec m (blockOf p)).
    destruct (WSE.Miff.In_dec (offsetOf p) (blockOffs0 m (blockOf p))).
    auto.
    right; intros [X Y]; auto.
    right; intros [X Y]; auto.
  Qed.


  Lemma validPtr_dec : forall m p,
    {validPtr m p} + {~validPtr m p}.
  Proof.
    intros m p; apply (validPtr0_dec (mval m) p).
  Qed.


  (* constructors *)
  Definition empty0 :=
    mem0i NS.empty (const_map _ WS.empty) (const_map _ (empty_pmap _ _)).

  Lemma emptyOk : memOk empty0.
  Proof.
    unfold empty0; constructor; intros b; simpl.
    setsimpl; auto.
    intros _; msimpl.
    intros i NI; msimpl.
  Qed.

  Definition empty : mem := memi emptyOk.

  Definition sing_mem0 (p:addr) (v:val) : mem0 :=
    mem0i (NS.add (blockOf p) NS.empty)
    (map_set eq_nat_dec (const_map _ WS.empty) 
      (blockOf p) (WS.add (offsetOf p) WS.empty))
    (map_set eq_nat_dec (const_map _ (empty_pmap _ _)) (blockOf p)
      (pmap_set eq_word_dec (empty_pmap _ _) (offsetOf p) v)).

  Lemma sing_memOk : forall p v,
    memOk (sing_mem0 p v).
  Proof.
    unfold sing_mem0; constructor; intros b'; simpl;
      (destruct (eq_nat_dec (blockOf p) b') as [E | NE]; 
        [subst b' | ]; msimpl).
    
    intros _ E.
    assert (I : WS.In (offsetOf p) WS.empty); [ | eauto].
    rewrite <- E; setsimpl; auto.

    setsimpl in I; auto.

    setsimpl; intros I _ ; intuit.
    
    setsimpl; intros; intuit.

    intros i' NI; rewrite pmap_setNeq.
    msimpl.
    intro; subst i'; setsimpl in NI; intuit.

    intros i' _; msimpl.

  Qed.

  Definition sing_mem p v : mem := memi (sing_memOk p v).


  Fixpoint block_pmap' (n:nat) : word -> list val -> pmap word val :=
    fun i vv =>
      match n with
        | (0 | 1 | 2 | 3) => empty_pmap _ _
        | S (S (S (S n'))) => 
          (pmap_set eq_word_dec 
            (block_pmap' n' (i - 4) (tail vv)) (i - 4)
            (hd Vundef vv))%word
      end.

  Definition block_pmap (w:word) (vv:list val) :=
    block_pmap' (word_to_nat (wordAlign w)) (wordAlign w) vv.

  Definition block_mem0 (b:block) (i:word) (vv:list val) :=
    mem0i 
    (if eq_word_dec (wordAlign i) 0 then 
      NS.empty else NS.add b NS.empty)
    (map_set eq_nat_dec (const_map _ WS.empty) b (zeroWordSet i))
    (map_set eq_nat_dec (const_map _ (empty_pmap _ _)) b (block_pmap i vv)).

  Lemma block_memOk : forall b i vv,
    memOk (block_mem0 b i vv).
  Proof.
    unfold block_mem0; intros b i vv; constructor; intros b'; simpl.

    intros I; destruct (eq_word_dec (wordAlign i) 0) as [E | NE].
    setsimpl in I; intuit.

    intro E; setsimpl in I; intuit; subst; msimpl.
    assert (I : WS.In 0 WS.empty).
    rewrite <- E; clear E; setsimpl.
    constructor.
    apply wordInj; simpl; rewrite wordModEq; wauto.
    generalize dependent (wordAlign i).
    intros; hyr NE (wordNeInj NE); destruct w; wunfold; omega.
    setsimpl in I; auto.

    destruct (eq_word_dec (wordAlign i) 0%word) as [E | NE]; setsimpl.

    intros _; destruct (eq_nat_dec b b') as [E' | NE]; subst; msimpl.
    apply zeroWordSetEmptyEq; auto.

    intro I.
    assert (NE' : b <> b').
    intro E; subst; auto.
    msimpl.

    destruct (eq_nat_dec b b') as [E | NE].
    2:intros i' _; msimpl.
    subst b'; msimpl; clear b; intros i' I'; setsimpl in I'.
    
    unfold block_pmap; generalize dependent vv; generalize dependent i'.
    induction i using wordInd4; simpl; intros; msimpl.
    rewrite wordAlignAddEq; [ | auto]; rewrite wordAddNatEq.
    replace (word_to_nat 4) with 4%nat.
    replace (word_to_nat (wordAlign i) + 4)%nat with 
      (4 + word_to_nat (wordAlign i))%nat by auto with * ; simpl.
    rewrite wordAddSubEq.
    destruct (eq_word_dec (wordAlign i) i') as [E | NE]; subst; msimpl.
    elf; clear IHi; apply I'; clear I'.
    rewrite wordAlignAlignEq; constructor; auto.
    apply wordAlignLtInj; apply wordAddLtI; auto.
    apply wordLeInj'; omega.
    apply wordNeInj'; wunfold; omega.
    apply IHi; clear IHi; intros [E Lt]; apply I'; clear I'; constructor; auto.
    rewrite E in Lt; rewrite E.
    rewrite wordAlignAddEq.
    apply wordAddLtI.
    wunfold; omega.
    apply wordAlignLe; auto.
    apply wordNeInj'; wunfold; omega.
    auto.
    compute; auto.
    apply wordAlignLe; auto.
  Qed.


  Definition block_mem b w vv : mem := memi (block_memOk b w vv).


  (* operations *)
  Definition load (m:mem) (p:addr) : option val :=
    if validPtr0_dec (mval m) p then
      Some (match pmap_get (map_get (valsOf (mval m)) (blockOf p))
              (offsetOf p) with
              | Some v => v
              | None => Vundef
            end)
    else
      None.

  Definition store0 (m:mem0) (p:addr) (v:val) : option mem0 :=
    if validPtr0_dec m p then
      Some (mem0i (validOf m) (addrsOf m)
        (map_set eq_nat_dec (valsOf m) (blockOf p)
          (pmap_set eq_word_dec (map_get (valsOf m) (blockOf p))
            (offsetOf p) v)))
    else
      None.


  Definition mvalOf (mo:option mem0) : mem0 :=
    match mo with
      | Some m => m
      | None => empty0
    end.

  Lemma storeOk : forall m,
    memOk m -> forall p v,
      memOk (mvalOf (store0 m p v)).
  Proof.
    unfold store0; intros m mOk p v.
    destruct (validPtr0_dec m p) as [V | _].
    2:apply emptyOk.
    destruct mOk as [M1 M2 M3].
    constructor; intros b'; simpl; auto.

    intros i' NI.
    destruct (eq_block_dec (blockOf p) b') as [E | NE]; subst; msimpl.
    destruct (eq_word_dec (offsetOf p) i') as [E | NE]; subst; msimpl.
    destruct V; elf; auto.

  Qed.


  Definition store m p v : option mem :=
    match store0 (mval m) p v with
      | Some _ => Some (memi (storeOk (mok m) p v))
      | None => None
    end.

  Definition free0 m b :=
    if validBlock0_dec m b then
      Some (mem0i (NS.remove b (validOf m))
        (map_set eq_nat_dec (addrsOf m) b WS.empty)
        (map_set eq_nat_dec (valsOf m) b (empty_pmap _ _)))
    else
      None.

  Lemma freeOk : forall m,
    memOk m -> forall b,
      memOk (mvalOf (free0 m b)).
  Proof.
    unfold free0; intros m mOk b.
    destruct (validBlock0_dec m b) as [V | _].
    2:apply emptyOk.
    destruct mOk as [M1 M2 M3]; simpl; constructor; simpl; intros b'.

    intro I; setsimpl in I; intuit; msimpl; eauto.

    intros NI; destruct (eq_block_dec b b') as [E | NE].
    subst b'; msimpl.
    msimpl.
    apply M2; setsimpl in NI; setsimpl; intuit.

    destruct (eq_block_dec b b') as [E | NE].
    intros i _; subst b'; msimpl.
    msimpl.

  Qed.

  Definition free m b : option mem :=
    match free0 (mval m) b with
      | Some _ => Some (memi (freeOk (mok m) b))
      | None => None
    end.


  (* utility functions *)

  Lemma injHelper : forall (A B:Type) (M:option A) (X Y:B),
    match M with
      | Some _ => Some X
      | None => None
    end = Some Y ->
    M <> None /\ X = Y.
  Proof.
    intros A B M X Y E; destruct M.
    injection E; intros; constructor; auto; discriminate.
    discriminate E.
  Qed.

  Lemma storeInj : forall m p v m',
    store m p v = Some m' ->
    store0 (mval m) p v = Some (mval m').
  Proof.
    unfold store; intros [m mOk] p v [m' m'Ok] E; simpl in * .
    destruct (injHelper (store0 m p v) _ E) as [NE E']; clear E.
    injection E'; clear E'; intro; subst.
    destruct (store0 m p v); auto; elf; auto.
  Qed.
  Implicit Arguments storeInj [m p v m'].


  Lemma freeInj : forall m b m',
    free m b = Some m' ->
    free0 (mval m) b = Some (mval m').
  Proof.
    unfold free; intros [m mOk] b [m' m'Ok] E; simpl in * .
    destruct (injHelper (free0 m b) _ E) as [NE E']; clear E.
    injection E'; clear E'; intro; subst.
    destruct (free0 m b); intuition.
  Qed.
  Implicit Arguments freeInj [m b m'].

  Lemma introHelperSome : forall A B (X:option A) U (Y:B),
    X = Some U ->
    match X with
      | Some _ => Some Y
      | None => None
    end = Some Y.
  Proof.
    intros A B X U Y E; rewrite E; auto.
  Qed.

  Lemma introHelperNone : forall A B (X:option A) (Y:B),
    X = None ->
    match X with
      | Some _ => Some Y
      | None => None
    end = None.
  Proof.
    intros A B X Y E; rewrite E; auto.
  Qed.

  Lemma freeI : forall m b m',
    free0 (mval m) b = Some m' ->
    free m b = Some (memi (freeOk (mok m) b)).
  Proof.
    unfold free; intros [m mOk] b m' E; simpl in * .
    apply (@introHelperSome mem0 mem (free0 m b) m'); auto.
  Qed.

  Lemma storeI : forall m p v m',
    store0 (mval m) p v = Some m' ->
    store m p v = Some (memi (storeOk (mok m) p v)).
  Proof.
    unfold store; intros [m mOk] p v m' E; simpl in * .
    apply (@introHelperSome mem0 mem (store0 m p v) m'); auto.
  Qed.
  Implicit Arguments storeI [m p v m'].


  (* validity properties *)
  
  Lemma emptyValidE : forall b, 
    validBlock empty b -> 
    False.
  Proof.
    unfold validBlock, validBlock0; simpl; intros b V; setsimpl in V; intuit.
  Qed.

  Lemma sing_memValidI : forall p v,
    validBlock (sing_mem p v) (blockOf p).
  Proof.
    unfold validBlock, validBlock0; simpl; intros p v; setsimpl; intuit.
  Qed.

  Lemma sing_memValidE : forall p b v,
    validBlock (sing_mem p v) b ->
    b = blockOf p.
  Proof.
    unfold validBlock, validBlock0; simpl; intros p b v V; setsimpl in V; intuit.
  Qed.

  Lemma block_memValidI : forall b i vv,
    0%word <> wordAlign i ->
    validBlock (block_mem b i vv) b.
  Proof.
    unfold validBlock, validBlock0; simpl; intros b i vv LE.
    destruct (eq_word_dec (wordAlign i) 0%word) as [E | NE]; setsimpl; auto.
  Qed.

  Lemma block_memValidE : forall b b' i vv,
    validBlock (block_mem b i vv) b' ->
    b = b' /\ 0%word <> wordAlign i.
  Proof.
    unfold validBlock, validBlock0; simpl; intros b b' i vv V.
    destruct (eq_word_dec (wordAlign i) 0%word); setsimpl in V; intuit.
  Qed.


  Lemma storeValidI : forall m m' p v,
    store m p v = Some m' -> forall b',
      validBlock m' b' <-> validBlock m b'.
  Proof.
    unfold validBlock, validBlock0; simpl; intros [m mOk] [m' m'Ok] p v E b'.
    hyr E (storeInj E); simpl in *; clear mOk m'Ok; unfold store0 in E.
    destruct (validPtr0_dec m p) as [V | NV].
    injector E; tauto.
    discriminate.
  Qed.

  Lemma storeValidI1 : forall m m' p v,
    store m p v = Some m' -> forall b',
      validBlock m' b' -> validBlock m b'.
  Proof.
    intros m m' p v E b' V; destruct (storeValidI _ _ _ E b'); auto.
  Qed.
  
  Lemma storeValidI2 : forall m m' p v,
    store m p v = Some m' -> forall b',
      validBlock m b' -> validBlock m' b'.
  Proof.
    intros m m' p v E b' V; destruct (storeValidI _ _ _ E b'); auto.
  Qed.


  Lemma freeValidI1 : forall m m' b,
    free m b = Some m' -> forall b',
      validBlock m' b' -> 
      b' <> b /\ validBlock m b'.
  Proof.
    unfold validBlock, validBlock0; intros [m mOk] [m' m'Ok] b E b' V.
    hyr E (freeInj E); simpl in *; clear mOk m'Ok; unfold free0 in E.
    destruct (validBlock0_dec m b) as [V' | _].
    injector E; simpl in * ; setsimpl in V; intuit.
    discriminate.
  Qed.

  Lemma freeValidI2 : forall m m' b,
    free m b = Some m' -> forall b',
      b' <> b -> 
      validBlock m b' -> 
      validBlock m' b'.
  Proof.
    unfold validBlock, validBlock0; intros [m mOk] [m' m'Ok] b E b' NE V.
    hyr E (freeInj E); simpl in *; clear mOk m'Ok; unfold free0 in E.
    destruct (validBlock0_dec m b) as [V' | _].
    injector E; simpl; setsimpl; intuit.
    discriminate.
  Qed.

  Lemma freeValidE : forall m b,
    validBlock m b ->
    exists m', free m b = Some m'.
  Proof.
    unfold validBlock, validBlock0; intros m b V; simpl in V.
    assert (E : exists m', free0 (mval m) b = Some m').
    unfold free0; destruct (validBlock0_dec (mval m) b).
    econstructor; auto.
    elf; auto.
    destruct E as [m' E]; rewrite (freeI _ _ E); econstructor; auto.
  Qed.

  Lemma freeInvalidE : forall m b,
    ~validBlock m b ->
    free m b = None.
  Proof.
    unfold free; intros [m mOk] b NV; simpl; apply introHelperNone.
    unfold free0; destruct (validBlock0_dec m b) as [V | _]; auto; elf; auto.
  Qed.

  Lemma invalidBlockEmptyI : forall m b,
    ~validBlock m b ->
    blockOffs m b = WS.empty.
  Proof.
    unfold validBlock, validBlock0, blockOffs, blockOffs0.
    intros [m [M1 M2 M3]] b NV; simpl in * ; auto.
  Qed.
 
  Lemma invalidBlockEmptyE : forall m b,
    blockOffs m b = WS.empty ->
    validBlock m b ->
    False.
  Proof.
    unfold validBlock, validBlock0, blockOffs, blockOffs0.
    intros [m [M1 M2 M3]] b E V; simpl in * .
    hyr M1 (M1 _ V); auto.
  Qed.


  (** blockOffs properties *)

  Lemma sing_memAddrsEq : forall p v,
    blockOffs (sing_mem p v) (blockOf p) = WS.add (offsetOf p) WS.empty.
  Proof.
    unfold blockOffs, blockOffs0, sing_mem; simpl; intros p v; msimpl.
  Qed.

  Lemma block_memAddrsEq : forall b i vv,
    blockOffs (block_mem b i vv) b = zeroWordSet i.
  Proof.
    unfold blockOffs, blockOffs0, sing_mem; simpl; intros b i v; msimpl.
  Qed.

  Lemma block_memAddrsNeq : forall b b',
    b <> b' -> forall len vv,
      blockOffs (block_mem b len vv) b' = WS.empty.  
  Proof.
    unfold blockOffs, blockOffs0, sing_mem; simpl;
      intros b b' NE i vv; msimpl.
  Qed.

  Lemma storeAddrsEq : forall m m' p v,
    store m p v = Some m' -> forall b',
      blockOffs m' b' = blockOffs m b'.
  Proof.
    unfold blockOffs, blockOffs0.
    intros [m mOk] [m' m'Ok] p v E; hyr E (storeInj E); simpl in * .
    unfold store0 in E.
    destruct (validPtr0_dec m p) as [V | _].
    injection E; clear E; intro; subst.
    simpl; auto.
    discriminate.
  Qed.

  Lemma freeAddrsEq : forall m m' b,
    free m b = Some m' -> forall b',
      b' <> b -> 
      blockOffs m' b' = blockOffs m b'.
  Proof.
    unfold blockOffs, blockOffs0.
    intros [m mOk] [m' m'Ok] b E b' NE; hyr E (freeInj E); simpl in * .
    unfold free0 in E; destruct (validBlock0_dec m b) as [V | _].
    injection E; clear E; intro; subst; simpl; msimpl.
    discriminate.
  Qed.


  (* properties of valid pointers *)

  Lemma emptyPtrE : forall p,
    validPtr empty p ->
    False.
  Proof.
    intros p [V _]; apply (emptyValidE V).
  Qed.

  Lemma sing_memPtrI : forall p v,
    validPtr (sing_mem p v) p .
  Proof.
    intros p v; apply validPtrEqv2; simpl; unfold sing_mem0; constructor; 
      unfold validBlock0, blockOffs0; simpl; msimpl; setsimpl; intuit.
  Qed.

  Lemma sing_memPtrE : forall p p' v,
    validPtr (sing_mem p v) p' ->
    p' = p.
  Proof.
    intros [b k] [b' k'] v V; hyr V (validPtrEqv1 V); destruct V as [V V'].
    simpl in * ; unfold validBlock0 in V; unfold blockOffs0 in V'.
    simpl in V, V'; autorewrite with SetEq in V; intuition; subst; 
      msimpl; setsimpl in V'; intuit; subst; auto.
  Qed.

  Lemma block_memPtrI : forall p i,
    WS.In (offsetOf p) (zeroWordSet i) -> forall vv,
    validPtr (block_mem (blockOf p) i vv) p.
  Proof.
    intros p i I vv; apply validPtrEqv2.
    simpl; constructor.
    unfold validBlock0; simpl.
    destruct (eq_word_dec (wordAlign i) 0%word) as [E | NE].
    subst; rewrite zeroWordSetEmptyEq in I; auto; elf; setsimpl in I; auto.
    setsimpl; intuit.
    unfold blockOffs0; simpl; msimpl.
  Qed.

  Lemma block_memPtrE : forall b i p vv,
    validPtr (block_mem b i vv) p ->
    blockOf p = b /\ WS.In (offsetOf p) (zeroWordSet i).
  Proof.
    intros b i p vv V; hyr V (validPtrEqv1 V); destruct V as [V V'].
    unfold validBlock0 in V; unfold blockOffs0 in V'; simpl in * .
    destruct (eq_word_dec (wordAlign i) 0%word) as [E | NE].
    setsimpl in V; intuit.
    cut (blockOf p = b).
    intro; subst; msimpl.
    clear V'; setsimpl in V; intuit.
  Qed.


  Lemma storePtrI1 : forall m m' p v,
    store m p v = Some m' -> forall p',
      validPtr m' p' -> validPtr m p'.
  Proof.
    intros [m mOk] [m' m'Ok] p v E p' V; hyr E (storeInj E).
    hyr V (validPtrEqv1 V); apply validPtrEqv2; simpl in * .
    unfold store0 in E; unfold validPtr0, validBlock0, blockOffs0 in * ;
      simpl in * .
    destruct (validPtr0_dec m p).
    injector E; auto.
    discriminate.
  Qed.


  Lemma storePtrI2 : forall m m' p v,
    store m p v = Some m' -> forall p',
      validPtr m p' -> validPtr m' p'.
  Proof.
    intros [m mOk] [m' m'Ok] p v E p' V; hyr E (storeInj E).
    hyr V (validPtrEqv1 V); apply validPtrEqv2; simpl in * .
    unfold store0 in E; unfold validPtr0, validBlock0, blockOffs0 in * ;
      simpl in * .
    destruct (validPtr0_dec m p).
    injector E; auto.
    discriminate.
  Qed.

  Lemma loadPtrI : forall m p v,
    load m p = Some v ->
    validPtr m p.
  Proof.
    unfold load; intros [m [M1 M2 M3]] p v E; simpl in E.
    apply validPtrEqv2; simpl.
    unfold validPtr0, blockOffs0, validBlock0.
    destruct (validPtr0_dec m p) as [V | _]; auto; discriminate.
  Qed.

  Lemma storePtrE : forall m p,
    validPtr m p -> forall v,
      exists m', store m p v = Some m'.
  Proof.
    intros m p V v.
    assert (E : exists m', store0 (mval m) p v = Some m').
    destruct m as [m mOk]; hyr V (validPtrEqv1 V); simpl in * .
    unfold store0; destruct (validPtr0_dec m p) as [_ | NV].
    econstructor; auto.
    elf; auto.
    destruct E as [m' E]; rewrite (storeI E); econstructor; auto.
  Qed.

  Lemma loadPtrE : forall m p,
    validPtr m p ->
    exists v, load m p = Some v.
  Proof.
    intros [m [M1 M2 M3]] p V; hyr V (validPtrEqv1 V); destruct V as [V V'].
    unfold load; simpl in * .
    unfold validBlock0 in V; unfold blockOffs0 in V'.
    destruct (validPtr0_dec m p) as [_ | NV].
    econstructor; auto.
    elf; apply NV; constructor; auto.
  Qed.

  Lemma invalidPtrE : forall m p,
    ~validPtr m p ->
    load m p = None.
  Proof.
    unfold load; intros [m mOk] p NI; simpl in * .
    destruct (validPtr0_dec m p) as [V | NV]; auto; elf; auto.
  Qed.
  

  (* basic operation properties *)

  Lemma sing_memLoadEq : forall p v,
    load (sing_mem p v) p = Some v.
  Proof.
    unfold load, sing_mem; intros p v; simpl.
    destruct (validPtr0_dec (sing_mem0 p v)).
    msimpl.
    elf; apply n; clear n.
    replace (sing_mem0 p v) with (mval (sing_mem p v)) by auto.
    apply validPtrEqv1; apply sing_memPtrI.
  Qed.

  Lemma memEqInj : forall x y,
    x = y -> forall (H:memOk x) (H':memOk y),
    memi H = memi H'.
  Proof.
    intros x y E H H'; subst y; rewrite (proof_irrel H H'); auto.
  Qed.

  Lemma sing_memStoreEq : forall p v v',
    store (sing_mem p v) p v' = Some (sing_mem p v').
  Proof.
    intros p v v'.
    assert (E : exists m', store0 (mval (sing_mem p v)) p v' = Some m').
    unfold store0; simpl.
    destruct (validPtr0_dec (sing_mem0 p v) p) as [_ | NV].
    econstructor; reflexivity.
    elf; apply NV.
    replace (sing_mem0 p v) with (mval (sing_mem p v)) by auto.
    apply validPtrEqv1; apply sing_memPtrI.
    destruct E as [m' E]; rewrite (storeI E); simpl; unfold sing_mem.
    f_equal; apply memEqInj.
    simpl in E; rewrite E; simpl.
    unfold store0 in E.
    destruct (validPtr0_dec (sing_mem0 p v) p) as [_ | NV].
    injection E; clear E; intro; subst; unfold sing_mem0.
    f_equal.
    apply map_ext; intros b'.
    destruct (eq_block_dec (blockOf p) b') as [E | NE]; subst; msimpl.
    apply pmap_ext; intros l'.
    destruct (eq_word_dec (offsetOf p) l') as [E | NE]; subst; msimpl.
    discriminate.
  Qed.

  Lemma storeLoadEq : forall m p v m',
    store m p v = Some m' ->
    load m' p = Some v.
  Proof.
    unfold load; intros [m mOk] p v [m' m'Ok] E; hyr E (storeInj E).

    simpl in * ; unfold store0 in E.
    destruct (validPtr0_dec m p).
    2:discriminate E.
    destruct (validPtr0_dec m' p) as [V | NV].
    injection E; clear E; intro; subst; simpl; msimpl.
    elf; apply NV; injector E; clear NV; destruct v0.
    unfold validBlock0, blockOffs0; constructor; auto.
  Qed.


  Lemma storeLoadNeq : forall m m' p v,
    store m p v = Some m' -> forall p',
      p <> p' ->
      load m' p' = load m p'.
  Proof.
    intros [m mOk] [m' m'Ok] p v E p' NE; hyr E (storeInj E).
    unfold store0 in E; simpl in * ; destruct (validPtr0_dec m p) as [V | NV].
    2:discriminate.
    unfold load; simpl in * .
    destruct (validPtr0_dec m' p') as [V' | NV']; (injector E; simpl in * );
      destruct (validPtr0_dec m p') as [V'' | NV''].

    assert (D : blockOf p <> blockOf p' \/ offsetOf p <> offsetOf p').
    destruct p; destruct p'.
    destruct (eq_block_dec b b0) as [E | NE'].
    destruct (eq_word_dec w w0) as [E' | NE']; subst; simpl; [elf |]; auto.
    simpl; auto.

    clear NE; destruct D as [NE | NE].
    msimpl.
    destruct (eq_nat_dec (blockOf p) (blockOf p')) as [E | NE'].
    rewrite <- E; msimpl.
    msimpl.
    elf; auto.
    elf; auto.
    auto.
  Qed.


  Lemma freeLoadNeq : forall m m' b,
    free m b = Some m' -> forall p,
      blockOf p <> b ->
        load m' p = load m p.
  Proof.
    intros [m mOk] [m' m'Ok] b E p NE; hyr E (freeInj E); unfold load;
      simpl in * .
    unfold free0 in E; destruct (validBlock0_dec m b) as [V | _].
    2:discriminate.
    destruct (validPtr0_dec m' p) as [V' | NV'];
      injector E; destruct (validPtr0_dec m p) as [V'' | NV'']; simpl in * .
    msimpl.
    elf; apply NV''; clear NV''.
    unfold validPtr0, validBlock0, blockOffs0 in V'; simpl in V'.
    destruct V' as [V1 V2].
    msimpl.
    assert (V1' := NS.remove_3 V1); constructor; auto.

    clear mOk m'Ok; elf.
    unfold validPtr0, validBlock0, blockOffs0 in * ; simpl in * .
    msimpl.
    destruct V'' as [V1 V2]; apply NV'; constructor; auto.
    setsimpl in NV'; intuit.
    auto.
  Qed.


  (* defining unions of memories *)

  Definition addr_pick p p' :=
    if eq_block_dec (blockOf p) (blockOf p') then
      if lt_word_dec (offsetOf p) (offsetOf p') then p else p'
    else if lt_dec (blockOf p) (blockOf p') then
      p
    else
      p'.

  Lemma addr_pickComm : forall b b',
    addr_pick b b' = addr_pick b' b.
  Proof.
    intros [b k] [b' k']; unfold addr_pick; simpl.
    destruct (eq_block_dec b b') as [E | NE].
    subst b'; destruct (eq_block_dec b b) as [_ | NE].
    2:elf; auto.
    destruct (lt_word_dec k k') as [LT | NLT];
      destruct (lt_word_dec k' k) as [LT' | NLT']; auto; wunfold.
    elf; omega.
    f_equal; apply wordInj; omega.
    destruct (eq_block_dec b' b) as [E | _].  elf; auto.
    destruct (lt_dec b b') as [X | Y];
      destruct (lt_dec b' b) as [X' | Y']; auto; elf; omega.
  Qed.


  Ltac assocCases :=
    match goal with
      | |- context [eq_block_dec ?b ?b] =>
        let X := fresh in
          destruct (eq_block_dec b b) as [_ | X]; [ | elf; auto]
      | H : ?b <> ?b' |- context [eq_block_dec ?b ?b'] =>
        destruct (eq_block_dec b b') as [X | _]; [elf; auto | auto]
      | H : context [lt_dec ?b ?b'] |- _ =>
        destruct (lt_dec b b'); auto
      | |- context [lt_dec ?b ?b'] =>
        destruct (lt_dec b b'); auto
      | |- context [lt_word_dec ?a ?a'] => 
        destruct (lt_word_dec a a'); auto
      | H : context [lt_word_dec ?a ?a'] |- _ => 
        destruct (lt_word_dec a a'); auto
    end.

  Lemma addr_pickAssoc : forall a a' a'',
    addr_pick (addr_pick a a') a'' = addr_pick a (addr_pick a' a'').
  Proof.
    unfold addr_pick, blockOf, offsetOf; intros [b k] [b' k'] [b'' k''].
    destruct (eq_block_dec b b') as [E | NE].
    subst b'; destruct (eq_block_dec b b'') as [E | NE].
    subst b''; repeat assocCases; elf; wunfold; omega.
    repeat assocCases;
      (destruct (eq_block_dec b b''); [subst | ]; auto; elf; omega).
    destruct (eq_block_dec b' b'') as [E | NE'].
    subst; repeat assocCases; elf; omega.
    repeat assocCases; try (elf; omega);
      (destruct (eq_block_dec b b''); [subst; auto | auto]); elf; omega.
  Qed.

  Definition val_pick (v v':val) : val :=
    match v, v' with
      | Vundef, _ => v'
      | _, Vundef => v
      | Vword w, Vword w' => 
        if lt_word_dec w w' then v else v'
      | Vword _, Vptr _ => v
      | Vptr _, Vword _ => v'
      | Vptr a, Vptr a' => Vptr (addr_pick a a')
    end.


  (* order of args doesn't affect val_pick *)
  Lemma val_pickComm : forall v v',
    val_pick v v' = val_pick v' v.
  Proof.
    destruct v; destruct v'; auto; simpl.

    destruct (lt_word_dec w w0) as [w1 | w1]; 
      destruct (lt_word_dec w0 w) as [w2 | w2]; auto.
    elf; wunfold; omega.
    destruct w as [w wOk]; destruct w0 as [w' w'Ok].
    assert (E : w = w').
    wunfold; omega.
    subst w'; rewrite (proof_irrel wOk w'Ok); auto.

    rewrite addr_pickComm; auto.

  Qed.


  Ltac val_pickSplitter :=
    match goal with
      | |- context [lt_word_dec ?v ?v'] => destruct (lt_word_dec v v')
      | |- context [eq_word_dec ?x ?y] => destruct (eq_word_dec x y); subst
    end;
    auto; try (elf; auto; fail); try (wunfold ; elf; omega).

  Lemma val_pickAssoc : forall v v' v'',
    val_pick (val_pick v v') v'' = val_pick v (val_pick v' v'').
  Proof.
    unfold val_pick; destruct v; destruct v'; destruct v''; auto;
      repeat val_pickSplitter.
    rewrite addr_pickAssoc; auto.
  Qed.


  Definition vunion (x y:option val) : option val :=
    match x, y with
      | None, _ => y
      | _, None => x
      | Some v, Some v' => Some (val_pick v v')
    end.

  Definition union0 (m m':mem0) : mem0 :=
    mem0i (NS.union (validOf m) (validOf m'))
      (map_union (fun s s' => s |_| s') (addrsOf m) (addrsOf m'))
      (map_union (pmap_union vunion) (valsOf m) (valsOf m')).


  Lemma emptyUnionEq : forall a b,
    a |_| b = WS.empty ->
    a = WS.empty.
  Proof.
    intros a b E; apply WSE.set_ext; intro x; constructor; intro I.
    rewrite <- E; setsimpl; intuit.
    setsimpl in I; intuit.
  Qed.


  Definition unionOk : forall m,
    memOk m -> forall m',
    memOk m' ->
    memOk (union0 m m').
  Proof.
    intros m [M1 M2 M3] m' [M1' M2' M3'].
    constructor; intros b; simpl.

    intros I; msimpl; intro E.
    setsimpl in I; intuit.
    apply (M1 _ H); apply (emptyUnionEq E).
    rewrite (fun a b => WSE.set_ext (WSE.Miff.union_sym a b)) in E.
    apply (M1' _ H).
    replace (WSE.M.union (map_get (addrsOf m') b) (map_get (addrsOf m) b)) with
      ((map_get (addrsOf m') b) |_| (map_get (addrsOf m) b)) in E by reflexivity.
      (* i don't know why we need to do this *)
    apply (emptyUnionEq E).

    msimpl; intro NI.
    rewrite M2.
    rewrite M2'.
    setsimpl; auto.
    intro; apply NI; setsimpl; intuit.
    intro; apply NI; setsimpl; intuit.
    
    msimpl; intros i NI; msimpl.
    clear M1 M1' M2 M2'.
    rewrite M3.
    rewrite M3'.
    auto.
    intro; apply NI; setsimpl; intuit.
    intro; apply NI; setsimpl; intuit.
  Qed.

  
  Definition union m m' : mem := memi (unionOk (mok m) (mok m')).


  (* equalities on unions of memories *)

  Lemma unionEmpty : forall m, 
    union m empty = m.
  Proof.
    unfold union; intros [[v len vv] mOk]; simpl; apply memEqInj.
    unfold union0; f_equal; simpl.

    setsimpl; auto.

    apply map_ext; intros b; msimpl; setsimpl; auto.

    apply map_ext; intros b; msimpl.
    apply pmap_ext; intros i; msimpl.
    unfold vunion; destruct (pmap_get (map_get vv b) i); auto.
  Qed.
    
  Lemma unionComm : forall m m',
    union m m' = union m' m.
  Proof.
    unfold union.
    intros [[v l vv] mOk] [[v' l' vv'] m'Ok]; simpl; apply memEqInj.
    unfold union0; simpl; f_equal.

    rewrite (fun a b => NSE.set_ext (NSE.Miff.union_sym a b)); auto.

    apply map_ext; intros b; msimpl.
    rewrite (fun a b => WSE.set_ext (WSE.Miff.union_sym a b)); auto.
    
    apply map_ext; intros b; msimpl.
    apply pmap_ext; intros i; msimpl.
    unfold vunion.
    destruct (pmap_get (map_get vv b) i);
      destruct (pmap_get (map_get vv' b) i); auto.
    rewrite val_pickComm; auto; apply addr_pickComm.

  Qed.


  Lemma unionAssoc : forall m m' m'',
    union (union m m') m'' = union m (union m' m'').
  Proof.
    intros [[v a x] X] [[v' a' x'] X'] [[v'' a'' x''] X''].
    unfold union; apply memEqInj; simpl; clear X X' X''.
    unfold union0; simpl; f_equal.

    rewrite (fun a b c => NSE.set_ext (NSE.Miff.union_assoc a b c)); auto.

    apply map_ext; intros b; msimpl.
    rewrite (fun a b c => WSE.set_ext (WSE.Miff.union_assoc a b c)); auto.

    apply map_ext; intros b; msimpl.
    apply pmap_ext; intros i; msimpl.
    destruct (pmap_get (map_get x b) i);
      destruct (pmap_get (map_get x' b) i);
        destruct (pmap_get (map_get x'' b) i); auto.
    simpl; rewrite val_pickAssoc; auto; apply addr_pickAssoc.
  Qed.


  (* properties of unioned memories  *)

  Lemma unionValidI : forall m b,
    validBlock m b -> forall m',
    validBlock (union m m') b.
  Proof.
    unfold validBlock, validBlock0; intros [m mOk] b V [m' m'Ok]; 
      simpl in * ; setsimpl; intuit.
  Qed.


  Lemma unionValidE : forall m m' b,
    validBlock (union m m') b ->
    validBlock m b \/ validBlock m' b.
  Proof.
    unfold validBlock, validBlock0; intros [m mOk] [m' m'Ok] b V;
      simpl in * ; setsimpl in V; intuit.
  Qed.

  Lemma unionBlockAddrsEq : forall m m' b,
    blockOffs (union m m') b = blockOffs m b |_| blockOffs m' b.
  Proof.
    unfold blockOffs, blockOffs0; intros [m mOk] [m' m'Ok] b; simpl; msimpl.
  Qed.

  Lemma unionPtrI : forall m p,
    validPtr m p -> forall m',
    validPtr (union m m') p.
  Proof.
    intros m p [V V'] m'; constructor.
    apply unionValidI; auto.
    rewrite unionBlockAddrsEq; setsimpl; intuit.
  Qed.

  Lemma unionPtrE : forall m m' p,
    validPtr (union m m') p ->
    validPtr m p \/ validPtr m' p.
  Proof.
    intros [[v l vv] [M1 M2 M3]] [[v' l' vv'] [M1' M2' M3']] p [bV iV].
    unfold validPtr, validBlock, blockOffs, validBlock0, blockOffs0 in * .
    simpl in * ; rewrite map_unionEq in iV.

    setsimpl in iV; setsimpl in bV.
    destruct iV as [I | I]; [left | right].

    constructor; intuit.
    destruct (NSE.Miff.In_dec (blockOf p) v) as [V | NV]; auto.
    rewrite M2 in I; auto; elf; setsimpl in I; auto.
    
    constructor; auto.
    destruct (NSE.Miff.In_dec (blockOf p) v') as [V | NV]; auto.
    rewrite M2' in I; auto; elf; setsimpl in I; intuit.

  Qed.


  (* domain disjointedness *)

  Definition dom_disj m m' :=
    (forall p, validPtr m p -> validPtr m' p -> False).

  Lemma block_memEmptyEq : forall b len vv,
    wordAlign len = 0%word ->
    block_mem b len vv = empty.
  Proof.
    unfold block_mem, empty; intros b len vv E; apply memEqInj.
    unfold block_mem0, empty0.
    rewrite E; destruct (eq_word_dec 0 0)%word as [_ | NE].
    2:elf; auto.
    f_equal; auto.

    setsimpl; apply map_ext; intro b'.
    msimpl; destruct (eq_block_dec b b') as [E' | NE]; subst; msimpl.
    apply zeroWordSetEmptyEq; auto.
    apply map_ext; intro b'; msimpl.
    destruct (eq_block_dec b b') as [E' | NE]; subst; msimpl.
    unfold block_pmap; rewrite E; auto.
  Qed.


  Lemma block_memZeroEq : forall b vv,
    block_mem b 0%word vv = empty.
  Proof.
    unfold block_mem, empty; intros b vv; apply memEqInj.
    unfold block_mem0, empty0.
    destruct (eq_word_dec (wordAlign 0%word) 0%word) as [_ | NE].
    2:elf; hyr NE (wordNeInj NE); auto.
    f_equal; auto.
    setsimpl; apply map_ext; intro b'.
    msimpl; destruct (eq_block_dec b b') as [E | NE]; subst; msimpl.
    apply map_ext; intro b'; msimpl.
    destruct (eq_block_dec b b') as [E | NE]; subst; msimpl.
  Qed.

  Lemma loadUnionEq : forall m p v,
    load m p = Some v -> forall m',
      dom_disj m m' ->
      load (union m m') p = Some v.
  Proof.
    unfold load; intros m p v E m' D.
    destruct (validPtr0_dec (mval m) p) as [V | NV].
    2:discriminate.
    injection E; clear E; intro; subst.
    destruct (validPtr0_dec (mval (union m m')) p) as [_ | NV'].
    2:elf; apply NV'; apply validPtrEqv1; apply unionPtrI; auto.
    f_equal.
    destruct m as [m mOk]; destruct m' as [m' m'Ok]; simpl.
    msimpl.
    hyr D (D p (validPtrEqv1 V)); simpl in V.
    unfold validPtr, validBlock, blockOffs in D; simpl in D.
    destruct m'Ok as [M1' M2' M3']; rewrite M3'.
    destruct (pmap_get (map_get (valsOf m) (blockOf p)) (offsetOf p)); auto.
    intro I; rewrite M2' in I; auto; setsimpl in I; auto.
  Qed.

  Lemma storeUnionEq : forall m m' p v,
    store m p v = Some m' -> forall m'',
      dom_disj m m'' ->
      store (union m m'') p v = Some (union m' m'').
  Proof.
    intros m m' p v E m'' D; hyr E (storeInj E).
    cut (store0 (mval (union m m'')) p v = Some (mval (union m' m''))).
    intro E'; rewrite (storeI E'); clear E.
    unfold union; simpl; f_equal; apply memEqInj.
    destruct m as [m mOk]; destruct m' as [m' m'Ok]; simpl in * .
    rewrite E'; auto.

    destruct m as [m mOk]; destruct m' as [m' m'Ok]; 
      destruct m'' as [m'' m''Ok]; unfold store0 in * .
    destruct (validPtr0_dec (mval (memi mOk)) p) as [V | _].
    2:discriminate E.
    injection E; clear E; intro; subst.
    destruct (validPtr0_dec (mval (union (memi mOk) (memi m''Ok))) p) as
      [_ | NV].
    2:elf; apply NV; apply validPtrEqv1; apply unionPtrI; auto.
    simpl in * ; f_equal; unfold union0; simpl; f_equal; auto.
    
    apply map_ext; intros b'; msimpl.
    destruct (eq_block_dec (blockOf p) b') as [E | NE].

    subst b'; msimpl.
    apply pmap_ext; intros i; msimpl.
    destruct (eq_word_dec (offsetOf p) i) as [E | NE].
    subst; msimpl.
    hyr D (D p V); unfold validPtr, validBlock, blockOffs in D; simpl in D.
    unfold validPtr0, validBlock0, blockOffs0 in *; simpl in * .
    destruct m''Ok as [M1 M2 M3]; rewrite M3; auto.
    intro I.
    destruct (NSE.Miff.In_dec (blockOf p) (validOf m'')) as [V' | NV']; auto.
    rewrite M2 in I; auto; setsimpl in I; auto.

    msimpl.

    msimpl.

  Qed.


  Lemma freeUnionEq : forall m m' b i vv,
    free (union m (block_mem b i vv)) b = Some m' ->
    ~validBlock m b ->
    m = m'.
  Proof.
    intros [m mOk] [m' m'Ok] b i vv E NV; hyr E (freeInj E).
    apply memEqInj; unfold validBlock in NV; simpl in * .
    unfold free0 in E.
    destruct (validBlock0_dec (union0 m (block_mem0 b i vv)) b) as [V | NV'].
    2:discriminate E.

    injector E; clear m'Ok.
    destruct m as [v len vs]; simpl; f_equal.

    destruct (eq_word_dec (wordAlign i) 0%word) as [E | NE].
    subst; setsimpl; apply NSE.set_ext; intros b'; constructor; intro I; setsimpl.
    constructor; auto; intro; subst; auto.
    setsimpl in I; destruct I; auto.
    apply NSE.set_ext; intros b'; constructor; intro I.
    setsimpl; intuit; subst; auto.
    setsimpl in I; intuit.


    apply map_ext; intros b'.
    destruct (eq_block_dec b b') as [E | NE].
    subst; msimpl.
    destruct mOk as [_ mOk _]; rewrite mOk; auto.

    msimpl; setsimpl; auto.

    apply map_ext; intros b'.
    destruct (eq_block_dec b b') as [E | NE]; subst; msimpl.
    apply pmap_ext; intros i'; msimpl.
    destruct mOk as [_ mOk mOk']; rewrite mOk'; auto.
    rewrite mOk; setsimpl; auto.

    apply pmap_ext; intros i'; msimpl.
    destruct (pmap_get (map_get vs b') i'); auto.

  Qed.


  Lemma block_pmapEndEq : forall i i' vv,
    (wordAlign i <= i')%word ->
    pmap_get (block_pmap i vv) i' = None.
  Proof.
    unfold block_pmap.
    induction i using wordInd4; try (simpl; intros; msimpl; fail).
    intros i' vv LE.
    rewrite wordAlignAddEq.
    rewrite wordAddNatEq.
    replace (word_to_nat 4%word) with 4%nat by (compute; auto).
    replace (word_to_nat (wordAlign i) + 4) with 
      (4 + word_to_nat (wordAlign i)) by auto with *; simpl.
    rewrite wordAddSubEq.
    destruct (eq_word_dec (wordAlign i) i') as [E | NE]; subst; msimpl.
    elf; clear IHi vv.
    assert (LT : (wordAlign i < wordAlign (i + 4))%word).
    clear LE; apply wordAlignLtInj.
    apply wordAddLtI; auto.
    wunfold; omega.
    apply wordNeInj'; wunfold; omega.
    wunfold; omega.

    apply IHi; clear IHi vv.
    cut (wordAlign i < i')%word.
    intro LT.
    hyr LT (wordLtInj LT); apply wordLeInj'; hyr NE (wordNeInj NE); 
      destruct i; destruct i'; wunfold; omega.
    apply anotherWordAlignLemma; auto.

    apply wordAlignLe; auto.

    auto.
    
  Qed.


  Lemma block_pmapTailEq : forall i v,
    (4 <= i + 4)%word ->
    pmap_get (block_pmap (i + 4)%word v) (wordAlign i) = Some (hd Vundef v).
  Proof.
    intros i v LE; unfold block_pmap.
    rewrite wordAlignAddEq.
    rewrite wordAddNatEq.
    replace (word_to_nat 4%word) with 4 by reflexivity.
    replace (word_to_nat (wordAlign i) + 4) with
      (4 + word_to_nat (wordAlign i)) by auto with * ; simpl.
    rewrite wordAddSubEq; msimpl.
    apply wordAlignLe; auto.
    auto.
  Qed.

  Lemma block_pmapPrevEq : forall i i',
    (4 <= i + 4)%word ->
    wordAlign i <> i' -> forall vv,
    pmap_get (block_pmap (i + 4)%word vv) i' =
    pmap_get (block_pmap i (tail vv)) i'.
  Proof.
    unfold block_pmap; intros i i' LE NE' vv.
    rewrite wordAlignAddEq.
    rewrite wordAddNatEq.
    replace (word_to_nat 4%word) with 4 by (compute; auto).
    replace (word_to_nat (wordAlign i) + 4) with 
      (4 + word_to_nat (wordAlign i)) by omega; simpl.
    rewrite wordAddSubEq; msimpl.
    apply wordAlignLe; auto.
    auto.
  Qed.


  (* you can split off the end of a non-empty block *)
  Lemma block_memSplit : forall w,
    (4 <= w + 4)%word -> forall b vv,
      block_mem b (w + 4)%word vv = 
      union (sing_mem (b, wordAlign w) (hd Vundef vv))
        (block_mem b w (tail vv)).
  Proof.
    unfold union, block_mem; intros w LE b vv; simpl; apply memEqInj.
    unfold union0, block_mem0; simpl; f_equal.

    destruct (eq_word_dec (wordAlign (w + 4))%word 0%word) as [E | _].
    assert (X := wordPlusLeNE LE); rewrite <- E in X; clear E; elf; apply X.
    rewrite wordAlignAddEq; auto.

    destruct (eq_word_dec (wordAlign w) 0%word) as [E | NE']; 
      subst; setsimpl; auto.
    apply NSE.set_ext; intros b'; constructor; intro I; setsimpl; setsimpl in I; intuit.

    apply map_ext; intros b'; msimpl.
    destruct (eq_block_dec b b') as [E | NE'].
    subst b'; msimpl; rewrite zeroWordSetSplitEnd; auto.
    apply WSE.set_ext; intro x; constructor; intro I; setsimpl; setsimpl in I; intuit.
    msimpl; setsimpl; auto.

    apply map_ext; intros b'; msimpl.
    destruct (eq_block_dec b b') as [E | NE']; subst; msimpl.
    apply pmap_ext; intros i'; msimpl.
    destruct (eq_word_dec (wordAlign w) i') as [E | NE']; subst; msimpl.
    rewrite block_pmapTailEq; auto.
    rewrite block_pmapEndEq; auto; wunfold; omega.

    simpl; apply block_pmapPrevEq; auto.

    msimpl; apply pmap_ext; intros i'; msimpl.

  Qed.


End CM.


Export CM.



Implicit Arguments storeLoadEq [m p v m'].
Implicit Arguments storeLoadNeq [m m' p v].
Implicit Arguments storeValidI1 [m m' p v].
Implicit Arguments storeValidI2 [m m' p v].
Implicit Arguments storePtrE [m p].
Implicit Arguments storeAddrsEq [m m' p v].
Implicit Arguments storePtrI1 [m m' p v].
Implicit Arguments storePtrI2 [m m' p v].
Implicit Arguments emptyPtrE [p].
Implicit Arguments loadPtrI [m p v].
Implicit Arguments storeUnionEq [m m' p v].
Implicit Arguments block_memSplit [w].




