(******************************************************************

   Separation logic toolkit

   Andrew McCreight
   HASP Project
   Portland State University

   http://hasp.cs.pdx.edu/

   Copyright 2009 Portland State University

   This file is part of the separation logic toolkit.

   The separation logic toolkit is free software: you can redistribute
   it and/or modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation, either
   version 3 of the License, or (at your option) any later version.

   The separation logic toolkit is distributed in the hope that it
   will be useful, but WITHOUT ANY WARRANTY; without even the implied
   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with the separation logic toolkit.  If not, see
   <http://www.gnu.org/licenses/>.

*******************************************************************)


Require Import ProgLog.
Require Import ZArith.
Require Import ListAddrSet.
Require Export Array.


Set Implicit Arguments.


Section MC.


Open Scope word_scope.
Open Scope addr_scope.


(* defining the procedure *)

Definition lenv := Var 4.
Definition srcv := Var 5.
Definition dstv := Var 6.
Definition iv := Var 7.

(* arguments: srcv, destv, lenv
   return: nothing 
*)
Definition memCopyBody := (
  iv <- 0;;
  while (iv < lenv) (
    [dstv + 4 * iv] <- [srcv + 4 * iv];;
    iv <- iv + 1
  );;
  ret null_ptr
)%CM.


Definition memCopyDef :=
  fdefi (srcv::dstv::lenv::nil) (iv::nil) 0 memCopyBody.


(* only allows non-overlapping buffers *)
Definition memCopyPre (src dst:addr) l (v1 v2 v3:val) :=
  lexists l',
    !(v1 = src /\ v2 = dst /\ v3 = length l /\
      length l = length l') **
    array src l ** array dst l'.

Definition memCopyPost src dst l (_ _ _ v:val) :=
  array src l ** array dst l.

Notation memCopyTy := (addr::addr::list val::nil : tslist).

Definition memCopySpec := speci memCopyTy 3%nat memCopyPre memCopyPost.


Definition inv l src dst s :=
  exists l0, exists ls, exists ld,
    let len := length l in
    let i := length l0 in
    (array src l0 ** array dst l0 **
      array (src + 4 * i) ls ** array (dst + 4 * i) ld) (cmemOf s) /\
    vfEqv (srcv::dstv::lenv::iv::nil) 
      ((srcv, src:val)::(dstv, dst:val)::
        (lenv, (len:val))::(iv, (i:val))::nil) (cvfOf s) /\
    l = l0 ++ ls /\
    length ls = length ld.


Ltac mcsimpl := autorewrite with list AddrSimpl ProgLog.


Lemma memCopyOk : forall P,
  fdefOk P memCopyTy 3%nat memCopyPre memCopyDef memCopyPost.
Proof.
  intros P; unfold memCopyDef, memCopyBody, memCopyPre, memCopyPost.
  fdefBegin.
  intros src dst l v1 v2 v3 m sp Hp vf VFE; ssimpl in Hp.
  subst; simpl in VFE. rename x into l'; rename H4 into LE.
  vcSteps.
  exists (inv l src dst).
  constructor.

  (* base case of loop *)
  unfold inv.
  exists (@nil val); exists l; exists l'.
  mcsimpl. intuition. sarewriteI.

  (* inductive case *)
  clear - LE; intros [m vf sp] [l0 [ls [ld [Hp [VFE [El Elen]]]]]].
  vcSteps. branchStep.

    (* case: still in the loop *)
    (* first, show that neither list is empty *)
    destruct ls as [| v ls].
      (* subcase: ls = nil *)
      subst; clear - E; elf.
      autorewrite with list in E. womega.

      (* subcase: ls <> nil *)
      (* ld is also not nil *)
      destruct ld. discriminate Elen.
      simpl in Elen; injector Elen.
      sarewriteE in Hp.
      mcsimpl. vcSteps.
      (* show that we've reestablished the loop invariant *)
      exists (l0 ++ (v ::nil)); exists ls; exists ld.
      cbv zeta; mcsimpl.
      replace (4 * (length l0 + 1))%word with (4 * length l0 + 4)%word by ring.
      mcsimpl. intuition.
      sarewriteI. searchMatch.
      vfEqvSolver; mcsimpl; auto.

    (* exit case *)
      ssimpl.
    (* show that the list we have yet to copy is empty *)
    assert (El0 : length ls = 0%nat).
      subst; clear - E Hp.
      (* show that neither length overflows *)
      assert (X1 := arrayLenWordBounded Hp).
      scl3 Hp; scl2 Hp; scar Hp. sarewriteI in Hp.
      assert (X2 := arrayLenWordBounded Hp).
      clear - E X1 X2.
      autorewrite with list in *. womega.
    destruct ls. 2:discriminate.
    destruct ld. 2:discriminate.
    clear - Hp El.
    autorewrite with list in El; subst l0.
    sarewriteE in Hp. auto.
  Qed.

End MC.