(******************************************************************

   Separation logic toolkit

   Andrew McCreight
   HASP Project
   Portland State University

   http://hasp.cs.pdx.edu/

   Copyright 2009 Portland State University

   This file is part of the separation logic toolkit.

   The separation logic toolkit is free software: you can redistribute
   it and/or modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation, either
   version 3 of the License, or (at your option) any later version.

   The separation logic toolkit is distributed in the hope that it
   will be useful, but WITHOUT ANY WARRANTY; without even the implied
   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with the separation logic toolkit.  If not, see
   <http://www.gnu.org/licenses/>.

*******************************************************************)


Require Export ListWordSet.
Require Import MemorySig.
Require Import CoreMemory.
Require Import ListNatSet.


Set Implicit Arguments.


Module M : MemorySig.

  Module CM := CoreMemory.CM.

  Inductive mem0 : Type :=
    mem0i : block -> NS.t -> CM.mem -> mem0.
  
  Definition alOf (m:mem0) := let (b, _, _) := m in b.
  Definition valOf (m:mem0) := let (_, v, _) := m in v.
  Definition coreOf (m:mem0) := let (_, _, c) := m in c.

  Inductive memOk (m:mem0) : Prop := 
    memOki : 
    (forall b, alOf m <= b -> ~NS.In b (valOf m)) ->
    (forall b, CM.validBlock (coreOf m) b -> NS.In b (valOf m)) ->
    (forall b, NS.In b (valOf m) -> CM.blockOffs (coreOf m) b <> WS.empty ->
      CM.validBlock (coreOf m) b) ->
    memOk m.

  Record mem' : Type := memi {mval : mem0; mok : memOk mval}.
  Definition mem := mem'.

  Notation block := CM.block.


  (* predicates *)
  Definition validBlock (m:mem) (b:block) : Prop := NS.In b (valOf (mval m)).
  Definition blockOffs (m:mem) : block -> WS.t := 
    CM.blockOffs (coreOf (mval m)).
  Definition validPtr (m:mem) (p:addr) : Prop :=
    validBlock m (blockOf p) /\ WS.In (offsetOf p) (blockOffs m (blockOf p)).


  (* constructor *)
  Definition empty0 := (mem0i 0 NS.empty CM.empty).

  Lemma emptyOk : memOk empty0.
  Proof.
    constructor; simpl.
    intros b _ V; setsimpl in V; auto.
    intros b V; elf; apply (emptyValidE V).
    intros b I _; elf; setsimpl in I; auto.
  Qed.

  Definition empty : mem := memi emptyOk.


  (* operations *)

  Definition alloc' (m:mem0) (len:word) : mem0 :=
    mem0i (S (alOf m)) (NS.add (alOf m) (valOf m)) 
      (union (coreOf m) (block_mem (alOf m) len nil)).

  Lemma allocOk : forall m,
    memOk m -> forall len,
    memOk (alloc' m len).
  Proof.
    intros m [ok1 ok2 ok3] len; constructor; simpl; intros b.

    intros LT I.
    compare b (alOf m); intros NE.
    subst b; omega.
    setsimpl in I; destruct I as [E | I']; auto.
    apply (ok1 b); auto; omega.

    intros V; destruct (unionValidE V) as [V' | V']; clear V.
    setsimpl; right; apply ok2; auto.
    destruct (block_memValidE V') as [E _]; subst b; setsimpl; auto.

    destruct (validBlock_dec (coreOf m) (alOf m)) as [V | NV].
    elf; apply (ok1 (alOf m)); auto.
    rewrite unionBlockAddrsEq.
    setsimpl; intros [I | I] NE.
    subst b; rewrite unionComm; apply CM.unionValidI; apply block_memValidI.
    intro E; apply NE; clear NE.
    rewrite block_memAddrsEq; rewrite zeroWordSetEmptyEq; auto; setsimpl.
    apply invalidBlockEmptyI; auto.

    destruct (eq_block_dec b (alOf m)) as [E | NE'].

    subst b.
    rewrite unionComm; apply CM.unionValidI; apply block_memValidI.
    intro E.
    rewrite block_memAddrsEq in NE.
    rewrite zeroWordSetEmptyEq in NE; auto; setsimpl in NE.
    apply NV; apply (ok3 (alOf m)); auto.

    apply CM.unionValidI; apply (ok3 b); auto; intro E.
    rewrite E in NE.
    rewrite (fun a b => WSE.set_ext (WSE.Miff.union_sym a b)) in NE.
    setsimpl in NE.
    rewrite block_memAddrsNeq in NE; auto.

  Qed.

  Definition alloc (m:mem) (len:word) : prod block mem :=
    (alOf (mval m), memi (allocOk (mok m) len)).

  Definition free' (m:mem0) (b:block) : option mem0 :=
    if NSE.Miff.In_dec b (valOf m) then
      Some (mem0i (alOf m) (NS.remove b (valOf m))
        (match CM.free (coreOf m) b with
           | Some c' => c'
           | None => coreOf m
         end))
    else
      None.

  
  Definition mvalOf (m:option mem0) : mem0 :=
    match m with
      | Some m' => m'
      | None => empty0
    end.

  Lemma freeOk : forall m,
    memOk m -> forall b,
      memOk (mvalOf (free' m b)).
  Proof.
    unfold free', mvalOf; intros m [ok1 ok2 ok3] b.
    destruct (NSE.Miff.In_dec b (valOf m)); [simpl | exact emptyOk].
    constructor; simpl; intros b'.

    intros LE I.
    setsimpl in I; destruct I as [NE I']; apply (ok1 b'); auto.

    assert (X : CM.free (coreOf m) b = None \/ 
      exists m', CM.free (coreOf m) b = Some m').
    destruct (CM.free (coreOf m) b); auto; right; econstructor; auto.

    destruct X as [E | [m' E]]; rewrite E; intro V.
    setsimpl.
    constructor.
    2:apply ok2; auto.
    intro; subst b'.
    destruct (CM.freeValidE V) as [m' E']; rewrite E in E'; discriminate E'.
    destruct (CM.freeValidI1 E V); setsimpl; constructor; auto; apply ok2; auto.

    assert (X : CM.free (coreOf m) b = None \/ 
      exists m', CM.free (coreOf m) b = Some m').
    destruct (CM.free (coreOf m) b); auto; right; econstructor; auto.

    intros I0; setsimpl in I0; destruct I0 as [NE I].
    destruct X as [E | [m' E]]; rewrite E; auto.

    intros NE'; apply (CM.freeValidI2 E); auto; apply ok3; auto.
    intros E'; apply NE'; clear NE'; rewrite (freeAddrsEq E); auto.

  Qed.


  Definition free m b : option mem :=
    match free' (mval m) b with
      | Some _ => Some (memi (freeOk (mok m) b))
      | None => None
    end.

  Definition load m p := CM.load (coreOf (mval m)) p.


  Definition store' (m:mem0) p (v:val) : option mem0 :=
    match CM.store (coreOf m) p v with
      | Some c' => Some (mem0i (alOf m) (valOf m) c')
      | None => None
    end.
  
  Lemma storeOk : forall m,
    memOk m -> forall p v,
      memOk (mvalOf (store' m p v)).
  Proof.
    unfold store'; intros m mOk p v.
    assert (D : CM.store (coreOf m) p v = None \/
      exists m', CM.store (coreOf m) p v = Some m').
    destruct (CM.store (coreOf m) p v); auto; right; econstructor; auto.
    destruct D as [E | [m' E]]; rewrite E.
    apply emptyOk.
    destruct mOk as [H H']; simpl; constructor; intro b'; simpl.
    intros LE I; apply (H b'); auto.
    intro V; apply (H' b'); apply (storeValidI1 E); auto.
    intros V NE; apply (storeValidI2 E); apply (H0 b'); auto.
    intros E'; apply NE; rewrite (storeAddrsEq E); auto.
  Qed.


  Definition store m p v : option mem :=
    match store' (mval m) p v with
      | Some _ => Some (memi (storeOk (mok m) p v))
      | None => None
    end.

  Definition core m := coreOf (mval m).


  (* decidability of predicates *)
  Definition eq_block_dec := CM.eq_block_dec.

  Lemma validBlock_dec : forall m b,
    {validBlock m b} + {~validBlock m b}.
  Proof.
    intros [m X] b; unfold validBlock; simpl; apply NSE.Miff.In_dec.
  Qed.

  Lemma validPtr_dec : forall m p,
    {validPtr m p} + {~validPtr m p}.
  Proof.
    intros m p; destruct (validBlock_dec m (blockOf p)) as [V | NV].
    destruct (WSE.Miff.In_dec (offsetOf p) (blockOffs m (blockOf p))) as [I | NI].
    left; constructor; auto.
    right; intros [_ X]; auto.
    right; intros [X _]; auto.
  Qed.


  (* validity properties *)
  
  Lemma emptyValidE : forall b, 
    validBlock empty b -> 
    False.
  Proof.
    unfold validBlock; simpl; intros b I; setsimpl in I; auto.
  Qed.

  Lemma coreValidE : forall m b,
    CM.validBlock (core m) b ->
    validBlock m b.
  Proof.
    unfold core, validBlock; intros [m mOk] b V; simpl in * . 
    destruct mOk as [_ X]; apply X; auto.
  Qed.

  Lemma allocFresh : forall m l b m',
    alloc m l = (b, m') ->
    ~validBlock m b.
  Proof.
    intros [m mOk] l b [m' m'Ok] E V.
    unfold validBlock in V; simpl in V.
    unfold alloc in E; simpl in E; injection E; clear E; intros; subst.
    destruct mOk as [mOk mOk']; apply (mOk (alOf m)); auto.
  Qed.

  Lemma allocValidI : forall m m' l b,
    alloc m l = (b, m') -> forall b',
      validBlock m' b' <-> b' = b \/ validBlock m b'.
  Proof.
    unfold alloc; intros [m mOk] [m' m'Ok] l b E b';
      simpl in E; injection E; clear E; intros; subst.
    unfold validBlock; simpl; constructor; intro I.
    setsimpl in I; auto.

    destruct I as [E | I].
    subst b'; setsimpl; auto.
    setsimpl; auto.
  Qed.

  Lemma allocValidI1 : forall m m' l b,
    alloc m l = (b, m') -> forall b',
      validBlock m' b' -> 
      b' = b \/ validBlock m b'.
  Proof.
    intros m m' l b E b' V; destruct (allocValidI E b'); auto.
  Qed.

  Lemma allocValidI2 : forall m m' l b,
    alloc m l = (b, m') -> forall b',
      b' = b \/ validBlock m b' ->
      validBlock m' b'.
  Proof.
    intros m m' l b E b' V; destruct (allocValidI E b'); auto.
  Qed.


  Lemma memOptInj : forall A (x:option A) m m',
    match x with Some _ => Some m | None => None end = Some m' ->
    mval m = mval m' /\ x <> None.
  Proof.
    destruct x; simpl; intros m m' E.
    injection E; intros; subst; constructor; auto; discriminate.
    discriminate E.
  Qed.
  Implicit Arguments memOptInj [A x m m'].


  Lemma storeValidI : forall m m' p v,
    store m p v = Some m' -> forall b',
      validBlock m' b' <-> validBlock m b'.
  Proof.
    unfold store; intros [m mOk] [m' m'Ok] p v E b'.
    destruct (memOptInj E) as [E' E'']; clear E; simpl in E', E''.
    unfold store' in * .
    assert (E : exists m', CM.store (coreOf m) p v = Some m').
    destruct (CM.store (coreOf m) p v); [econstructor | elf]; auto.
    destruct E as [m'' E]; rewrite E in E'; clear E''; simpl in E'; subst.
    unfold validBlock; simpl; constructor; auto.
  Qed.

  Lemma storeValidI1 : forall m m' p v,
    store m p v = Some m' -> forall b',
      validBlock m' b' -> validBlock m b'.
  Proof.
    intros m m' p v E b' V; destruct (storeValidI _ _ _ E b'); auto.
  Qed.
  
  Lemma storeValidI2 : forall m m' p v,
    store m p v = Some m' -> forall b',
      validBlock m b' -> validBlock m' b'.
  Proof.
    intros m m' p v E b' V; destruct (storeValidI _ _ _ E b'); auto.
  Qed.

  Lemma freeValidI : forall m m' b,
    free m b = Some m' -> forall b',
      (validBlock m' b' <-> (b' <> b /\ validBlock m b')).
  Proof.
    intros [m mOk] [m' m'Ok] b E b'.
    unfold free in E; destruct (memOptInj E) as [E' E'']; clear E.
    unfold validBlock; simpl in E', E''; simpl.
    unfold free' in * .
    destruct (NSE.Miff.In_dec b (valOf m)) as [I | NI].
    2:elf; auto.
    clear E''; unfold mvalOf in E'; subst m'; simpl.
    constructor; intro I'.
    setsimpl in I'; destruct I' as [NE I'']; auto.
    destruct I'; setsimpl; auto.
  Qed.


  Lemma freeValidI1 : forall m m' b,
    free m b = Some m' -> forall b',
      validBlock m' b' -> 
      b' <> b /\ validBlock m b'.
  Proof.
    intros m m' b E b' V; destruct (freeValidI _ _ E b'); auto.
  Qed.

  Lemma freeValidI2 : forall m m' b,
    free m b = Some m' -> forall b',
      b' <> b -> 
      validBlock m b' -> validBlock m' b'.
  Proof.
    intros m m' b E b' V; destruct (freeValidI _ _ E b'); auto.
  Qed.

  Lemma optionEsc : forall A (x:option A),
    x <> None -> forall B (a:B),
      match x with
        | Some _ => Some a 
        | None => None
      end = Some a.
  Proof.
    intros A x NE B a; destruct x; auto; elf; auto.
  Qed.


  Lemma freeValidE : forall m b,
    validBlock m b ->
    exists m', free m b = Some m'.
  Proof.
    unfold validBlock, free; intros [m mOk] b V; simpl in * .
    exists (memi (freeOk mOk b)); apply optionEsc; unfold free'.
    destruct (NSE.Miff.In_dec b (valOf m)); [discriminate | auto].
  Qed.


  (** blockOffs properties *)

  Lemma coreBlockAddrsEq : forall m b,
    blockOffs m b = CM.blockOffs (core m) b.
  Proof.
    unfold core, blockOffs; intros [m mOk] b; simpl; auto.
  Qed.


  Lemma allocAddrsEq : forall m m' l b,
    alloc m l = (b, m') ->
    blockOffs m' b = zeroWordSet l.
  Proof.
    unfold alloc, blockOffs; intros m m' l b E.
    assert (NV := allocFresh E).
    injection E; clear E; intros E E'; subst b m'; simpl.
    rewrite unionBlockAddrsEq.
    rewrite block_memAddrsEq.
    rewrite CM.invalidBlockEmptyI.
    rewrite (fun a b => WSE.set_ext (WSE.Miff.union_sym a b)); setsimpl; auto.
    intro V; apply NV; clear NV; apply coreValidE; auto.
  Qed.


  Lemma allocAddrsNeq : forall m m' l b,
    alloc m l = (b, m') -> forall b',
      b' <> b ->
      blockOffs m' b' = blockOffs m b'.
  Proof.
    unfold alloc, blockOffs; intros [m mOk] [m' m'Ok] l b E b' NE;
      simpl; simpl in E; injection E; clear E; intros; subst; simpl.
    rewrite unionBlockAddrsEq.
    replace (CoreMemory.CM.blockOffs (block_mem (alOf m) l nil) b')
      with WS.empty.
    setsimpl; auto.
    symmetry; apply CM.invalidBlockEmptyI; intro V.
    destruct (block_memValidE V); auto.
  Qed.
  
  Lemma storeAddrsEq : forall m m' p v,
    store m p v = Some m' -> forall b',
      blockOffs m' b' = blockOffs m b'.
  Proof.
    unfold store, blockOffs; intros [m mOk] [m' m'Ok] p v E b'.
    destruct (memOptInj E) as [E' E'']; clear E; simpl in * ; subst.
    unfold store' in * .
    assert (X : exists m', CM.store (coreOf m) p v = Some m').
    destruct (CM.store (coreOf m) p v); [econstructor | elf]; auto.
    destruct X as [m' E]; rewrite E; rewrite E in m'Ok; clear E''; simpl in * .
    rewrite (storeAddrsEq E); auto.
  Qed.


  Lemma freeAddrsEq : forall m m' b,
    free m b = Some m' -> forall b',
      b' <> b ->
      blockOffs m' b' = blockOffs m b'.
  Proof.
    unfold free, blockOffs; intros [m mOk] [m' m'Ok] b E b' NE.
    destruct (memOptInj E) as [E' E'']; clear E; simpl in * ; subst.
    unfold free' in * ; destruct (NSE.Miff.In_dec b (valOf m)).
    2:elf; auto.
    clear E''; simpl in * .
    assert (X : CM.free (coreOf m) b = None \/
      exists m', CM.free (coreOf m) b = Some m').
    destruct (CM.free (coreOf m) b); auto; right; econstructor; auto.
    destruct X as [E | [m' E]]; rewrite E; rewrite E in m'Ok; auto.
    apply (CM.freeAddrsEq E); auto.
  Qed.


  (* properties of valid pointers *)

  Lemma emptyPtrE : forall p,
    validPtr empty p ->
    False.
  Proof.
    intros p [V _]; apply (emptyValidE V).
  Qed.

  Lemma corePtrE : forall m p,
    CM.validPtr (core m) p ->
    validPtr m p.
  Proof.
    intros m p [V V']; constructor; auto; apply coreValidE; auto.
  Qed.

  Lemma allocPtrEqI : forall m m' l b,
    alloc m l = (b, m') -> forall i,
      (i < wordAlign l)%word ->
      validPtr m' (b, wordAlign i).
  Proof.
    intros m m' l b E i LT; constructor.
    apply (allocValidI2 E); auto.
    simpl; rewrite (allocAddrsEq E); setsimpl.
    constructor.
    rewrite wordAlignAlignEq; auto.
    assert (X := wordAlignLe' i).
    destruct l; destruct i; apply wordLtInj'; hyr X (wordLeInj X);
      hyr LT (wordLtInj LT); wunfold; omega.
  Qed.

  Lemma allocPtrNeqI : forall m m' l b,
    alloc m l = (b, m') -> forall p,
      blockOf p <> b ->
      validPtr m p ->
      validPtr m' p.
  Proof.
    intros m m' l b E p NE [V V']; constructor.
    apply (allocValidI2 E); auto.
    rewrite (allocAddrsNeq E); auto.
  Qed.

  Lemma storePtrI1 : forall m m' p v,
    store m p v = Some m' -> forall p',
      validPtr m' p' -> validPtr m p'.
  Proof.
    intros m m' p v E p' [V V']; constructor.
    apply (storeValidI1 _ _ _ E); auto.
    rewrite <- (storeAddrsEq _ _ _ E); auto.
  Qed.

  Lemma storePtrI2 : forall m m' p v,
    store m p v = Some m' -> forall p',
      validPtr m p' -> validPtr m' p'.
  Proof.
    intros m m' p v E p' [V V']; constructor.
    apply (storeValidI2 _ _ E); auto.
    rewrite (storeAddrsEq _ _ _ E); auto.
  Qed.

  Lemma freePtrI1 : forall m m' b,
    free m b = Some m' -> forall p,
      blockOf p <> b ->
	validPtr m' p -> validPtr m p.
  Proof.
    intros m m' b E p NE [V V']; constructor.
    destruct (freeValidI1 _ _ E V); auto.
    rewrite <- (freeAddrsEq _ E); auto.
  Qed.

  Lemma freePtrI2 : forall m m' b,
    free m b = Some m' -> forall p,
      blockOf p <> b ->
	validPtr m p -> validPtr m' p.
  Proof.
    intros m m' b E p NE [V V']; constructor.
    apply (freeValidI2 E); auto.
    rewrite (freeAddrsEq _ E); auto.
  Qed.

  Lemma loadPtrI : forall m p v,
    load m p = Some v ->
    validPtr m p.
  Proof.
    unfold load; intros [m mOk] p v E; simpl in E.
    assert (X := CM.loadPtrI E); destruct X as [V V']; clear E.
    constructor; auto.
    unfold validBlock; simpl; destruct mOk as [_ mOk]; apply mOk; auto.
  Qed.


  Lemma loadPtrE : forall m p,
    validPtr m p ->
    exists v, load m p = Some v.
  Proof.
    unfold load, validPtr, validBlock, blockOffs;
      intros [m mOk] p [V V']; simpl in * .
    assert (V'' : CM.validPtr (coreOf m) p).
    destruct mOk as [_ _ mOk]; hyr mOk (mOk _ V); constructor; auto.
    apply mOk; intro E; rewrite E in V'; setsimpl in V'; auto.
    exact (CM.loadPtrE V'').
  Qed.

  
  Lemma storePtrE : forall m p,
    validPtr m p -> forall v,
      exists m', store m p v = Some m'.
  Proof.
    unfold validPtr, validBlock, blockOffs, store.
    intros [m mOk] p [V1 V2] v.
    simpl in * .
    exists (memi (storeOk mOk p v)); apply optionEsc; unfold store'.
    assert (V' : CM.validPtr (coreOf m) p).
    destruct mOk as [_ _ mOk]; constructor; auto; apply mOk; auto.
    replace CM.blockOf with blockOf by auto.
    intro E; rewrite E in V2; setsimpl in V2; auto.
    destruct (storePtrE V' v) as [m' E].
    rewrite E; discriminate.
  Qed.


  (* basic operation properties *)

  Lemma allocLoadNeq : forall m m' l b,
    alloc m l = (b, m') -> forall p,
      blockOf p <> b ->
	load m' p = load m p.
  Proof.
    unfold load; intros [m mOk] [m' m'Ok] l b E p NE; simpl.
    assert (F := allocFresh E); unfold validBlock in F; simpl in F.
    unfold alloc in E; simpl in E; injection E; clear E; intros; subst.

    simpl; destruct (CM.validPtr_dec (coreOf m) p) as [V | NV].
    destruct (CM.loadPtrE V) as [v E]; rewrite E.
    rewrite (loadUnionEq E); auto.
    intros p3 [I1 _] [I2 _]; destruct (CM.block_memValidE I2); subst.
    destruct mOk as [_ mOk _]; rewrite H in F; auto.

    rewrite (CM.invalidPtrE NV); rewrite CM.invalidPtrE; auto; intros [I X].
    rewrite unionBlockAddrsEq in X; setsimpl in X; destruct X as [I' | I'].
    destruct (unionValidE I) as [I'' | I'']; clear I.
    apply NV; constructor; auto.
    apply NE; destruct (CM.block_memValidE I''); auto.
    rewrite CM.invalidBlockEmptyI in I'.
    setsimpl in I'; auto.
    intro V; apply NE; destruct (block_memValidE V); auto.
  Qed.


  Lemma freeLoadNeq : forall m m' b,
    free m b = Some m' -> forall p,
      blockOf p <> b ->
	load m' p = load m p.
  Proof.
    unfold free, load; 
      intros [m H] [m' H'] b E p NE; simpl in E; simpl.
    hyr E (memOptInj E); simpl in E; destruct E as [E NE'].
    subst m'; simpl; unfold free' in * .
    destruct (NSE.Miff.In_dec b (valOf m)) as [I | NI].
    2:elf; auto.
    clear NE'; simpl in * .
    assert (D : CM.free (coreOf m) b = None \/
      exists m', CM.free (coreOf m) b = Some m').
    destruct (CM.free (coreOf m) b); auto; right; econstructor; auto.
    destruct D as [E | [m' E]]; rewrite E; rewrite E in H'; auto.
    apply (CM.freeLoadNeq E); auto.
  Qed.

  
  Lemma storeLoadEq : forall m p v m',
    store m p v = Some m' ->
    load m' p = Some v.
  Proof.
    unfold load, store; intros [m mOk] p v [m' m'Ok] E; simpl in * .
    hyr E (memOptInj E); simpl in E; destruct E as [E NE].
    unfold store' in * .
    assert (D : CM.store (coreOf m) p v = None \/
      exists m', CM.store (coreOf m) p v = Some m').
    destruct (CM.store (coreOf m) p v); auto; right; econstructor; auto.
    destruct D as [E' | [m'' E']]; rewrite E' in E; rewrite E' in NE.
    elf; auto.
    subst; simpl in * ; rewrite (CM.storeLoadEq E'); auto.
  Qed.

  Lemma storeLoadNeq : forall m m' p v,
    store m p v = Some m' -> forall p',
      p <> p' ->
      load m' p' = load m p'.
  Proof.
    unfold load; intros [m mOk] [m' m'Ok] p v E p' D; simpl.
    hyr E (memOptInj E); simpl in E; destruct E as [E NE].
    unfold store' in * .
    assert (D' : CM.store (coreOf m) p v = None \/
      exists m', CM.store (coreOf m) p v = Some m').
    destruct (CM.store (coreOf m) p v); auto; right; econstructor; auto.
    destruct D' as [E' | [m'' E']]; rewrite E' in E; rewrite E' in NE.
    elf; auto.
    subst; simpl in * ; rewrite (CM.storeLoadNeq E'); auto.
  Qed.


  Lemma coreLoadEq : forall m p,
    CM.load (core m) p = load m p.
  Proof.
    unfold load; intros [m mOk] p; auto.
  Qed.

  Lemma coreStoreEq : forall m p v cm',
    CM.store (core m) p v = Some cm' ->
    exists m', store m p v = Some m' /\ core m' = cm'.
  Proof.
    unfold core, store; intros [m mOk] p v cm' E; simpl in * .
    econstructor.
    constructor.
    apply optionEsc; unfold store'; rewrite E; discriminate.
    simpl.
    unfold store'; rewrite E; auto.
  Qed.

  Lemma coreAllocEq : forall m m' i sp',
    alloc m i = (sp', m') ->
    exists vv, core m' = union (core m) (block_mem sp' i vv).
  Proof.
    unfold core, alloc; intros [m mOk] [m' m'Ok] i sp' E; simpl.
    exists (nil (A := val)).
    simpl in E; injection E; clear E; intros; subst; auto.
  Qed.


  Lemma freeCoreEq : forall m m' b,
    free m b = Some m' -> forall len cm' vv,
      core m = CM.union (CM.block_mem b len vv) cm' ->
      ~CM.validBlock cm' b ->
      core m' = cm'.
  Proof.
    unfold core, free; intros [m mOk] [m' m'Ok] b E len cm' vv Ceq NV; 
      simpl in * .
    hyr E (memOptInj E); simpl in E; destruct E as [E E']; subst; simpl.
    unfold free' in * .
    destruct (NSE.Miff.In_dec b (valOf m)) as [V | NV']; simpl in * .
    2:elf; auto.
    clear E'; destruct (CM.validBlock_dec (coreOf m) b) as [V' | NV'].
    destruct (CM.freeValidE V') as [m' E]; rewrite E; rewrite Ceq in E.
    rewrite unionComm in E; rewrite (CM.freeUnionEq E); auto.

    destruct (eq_word_dec (wordAlign len) 0) as [E | NE].

    rewrite block_memEmptyEq in Ceq; auto;
      rewrite unionComm in Ceq; rewrite unionEmpty in Ceq; subst cm'.
    rewrite freeInvalidE; auto.
    
    elf;  clear m'Ok.
    apply NV'; rewrite Ceq; clear NV' Ceq.
    apply unionValidI; apply block_memValidI; auto.

  Qed.



End M.


Export M.CM.
Export M.



(* properties of memory domain disjointedness *)


Lemma dom_disjI : forall m m',
  (forall p, CM.validPtr m p -> CM.validPtr m' p -> False) ->
  dom_disj m m'.
Proof.
  intros; auto.
Qed.


Lemma dom_disjE : forall m m',
  dom_disj m m' -> forall p,
    CM.validPtr m p ->
    CM.validPtr m' p ->
    False.
Proof.
  intros m m' D p V V'; apply (D p); auto.
Qed.


Lemma dom_disjEmpty : forall m,
  dom_disj m CM.empty.
Proof.
  intro m; apply dom_disjI; intros p V V'.
  apply (CM.emptyPtrE V').
Qed.


Lemma dom_disjComm : forall m m',
  dom_disj m m' ->
  dom_disj m' m.
Proof.
  intros m m' D; apply dom_disjI; intros p V V'; apply (D p); auto.
Qed.


Lemma dom_disjUnionI : forall m m',
  dom_disj m m' -> forall m'',
    dom_disj m m'' ->
    dom_disj m (union m' m'').
Proof.
  intros m m' D m'' D'; apply dom_disjI; intros p V V'.
  hyr D (D _ V); hyr D' (D' _ V); destruct (unionPtrE V'); clear V V'; auto.
Qed.


Lemma dom_disjUnionE : forall m m' m'',
  dom_disj (union m' m'') m ->
  dom_disj m' m.
Proof.
  intros m m' m'' D; intros p V V'; hyr D (D p); 
    apply (D (unionPtrI V _) V').
Qed.


(* store preserves disjointedness *)
Lemma dom_disjStore : forall m m' p v',
  CM.store m p v' = Some m' -> forall m'',
    dom_disj m m'' ->
    dom_disj m' m''.
Proof.
  intros m m' p v' E m'' D; intros p' I I'; apply (D p'); auto.
  clear I'; apply (CM.storePtrI1 E); auto.
Qed.
Implicit Arguments dom_disjStore [m m' p v'].


Lemma dom_disjBlock : forall w,
  (4 <= w + 4)%word -> forall b v vv,
    dom_disj (sing_mem (b, wordAlign w) v) (block_mem b w vv).
Proof.
  intros w LE b v vv; apply dom_disjI; intros p' [I V] [I' V'].
  assert (X := sing_memValidE I); clear I I'.
  rewrite X in V, V' .
  rewrite sing_memAddrsEq in V; rewrite block_memAddrsEq in V'.
  setsimpl in V; destruct V as [E | F].
  2:auto.
  subst; setsimpl in V'; destruct V' as [E' LT].
  destruct p'; simpl in * ; subst; wunfold; omega.
Qed.
Implicit Arguments dom_disjBlock [w].


Lemma eq_addr_dec : forall (x y:addr),
  {x = y} + {x <> y}.
Proof.
  intros [b i] [b' i'].
  destruct (eq_block_dec b b').
  subst b'; destruct (eq_word_dec i i').
  subst i'.
  left; reflexivity.
  right; intro X; injector X; auto.
  right; intro X; injector X; auto.
Qed.
