(******************************************************************

   Separation logic toolkit

   Andrew McCreight
   HASP Project
   Portland State University

   http://hasp.cs.pdx.edu/

   Copyright 2009 Portland State University

   This file is part of the separation logic toolkit.

   The separation logic toolkit is free software: you can redistribute
   it and/or modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation, either
   version 3 of the License, or (at your option) any later version.

   The separation logic toolkit is distributed in the hope that it
   will be useful, but WITHOUT ANY WARRANTY; without even the implied
   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with the separation logic toolkit.  If not, see
   <http://www.gnu.org/licenses/>.

*******************************************************************)



(* This is a minor variation of Reverse.v modified to be a little
   easier to follow. *)

Require Import ProgLog.
Require Import NatUtil.


Set Implicit Arguments.


Fixpoint llist (x:val) (l:list val) : mpred :=
  match l with
    | nil => !(x = 0%word)
    | v::l' =>
      lexists x', x |-> v ** (x + 4)%val |-> x' ** llist x' l'
  end.


Definition yv := Var 4.
Definition xv := Var 5.
Definition tv := Var 6.


Definition reverseBody := (
  yv <- 0;;
  while (xv != 0) (
    tv <- [xv + 4];;
    [xv + 4] <- yv;;
    yv <- xv;;
    xv <- tv
  );;
  ret yv
)%CM.


Notation reverseDef := (fdefi (xv::nil) (yv::tv::nil) 0 reverseBody).

Notation reverseTy := (list val::nil : tslist).

Definition reversePre is i :=
  llist i is.

Definition reversePost is (_ j:val) :=
  llist j (rev is).

Definition P0 : cdhpty := fun _ => None.

Definition inv l0 (s:cstate) :=
  exists w, exists v,
    (vfEqv (xv :: yv :: tv :: nil) ((yv,w) :: (xv, v) :: nil) (cvfOf s) /\
      exists vl, exists wl, 
        (llist v vl ** llist w wl) (cmemOf s) /\
        rev vl ++ wl = rev l0).


Lemma llist_non_null : forall x : val,
  x <> 0 -> forall l,
  mimpl (llist x l)
  (lexists h, lexists t, !(l = h::t) ** 
    lexists x', x |-> h ** (x + 4)%val |-> x' ** llist x' t).
Proof.
  intros x NE l m Hp.
  destruct l as [ | v l].

  simpl in Hp; ssimpl in Hp. elf; auto.

  unfold llist in Hp; fold llist in Hp. ssimpl in Hp.
  ssimpl; eauto.

Qed.


Lemma llist_cons : forall x v l,
  mimpl (llist x (v::l))
  (lexists y, x |-> v ** (x + 4)%val |-> y ** llist y l).
Proof.
  intros x v l m Hp; auto.
Qed.


Lemma llist_null : forall l,
  miff (llist 0%word l) (!(l = nil)).
Proof.
  intros l m; constructor; intros Hp; destruct l.

  simpl in Hp; ssimpl in Hp; ssimpl.

  simpl in Hp; ssimpl in Hp.
  (* why doesn't ssimpl in pick up the fact that the cont is on 0? *)
  elf; inversion Hp. inversion H0.

  simpl; ssimplg in Hp.

  ssimpl in Hp. discriminate H.

Qed.


Lemma llist_not_undef : forall l,
  mimpl (llist Vundef l) (!False).
Proof.
  intros l m Hp; destruct l; simpl in Hp; ssimpl in Hp.
  discriminate H.
Qed.



Hint Rewrite llist_non_null using assumption : SLogI.
Hint Rewrite llist_not_undef llist_null llist_cons : SLogI.


(* branchStep, except it doesn't do a vcSteps at the end, to make 
   the example a little clearer. *)

Ltac branchStep0X :=
  match goal with
    | |- context [valBoolCases ?P] =>
      let E := fresh "E" in
        destruct (valBoolCases P) as [[E | E] | E];
          autorewrite with valLogEq in E;
          autorewrite with valEq BlockOf in E;
          autorewrite with valEq2 valEq BlockOf in E;
            try discriminate E;
            try (elf; auto; fail)
  end.

Ltac branchStepX :=
  exists (None (A := spred));
    constructor; [branchStep0X | 
      let s := fresh in let X := fresh in intros s X; apply X].




Lemma reverseOk : fdefOk P0 reverseTy 1 reversePre reverseDef reversePost.
Proof.
  fdefBegin.
  unfold reversePre, reverseBody.
  intros vs x m sp Hp vf VFE. 
  ssimpl in Hp. simpl in VFE.
  vcSteps.
  exists (inv vs). split.

  (* base case *)
  econstructor; econstructor; intuition.
  exists vs. exists (@nil val).
  simpl. constructor. ssimpl. intuition.

  (* inductive case *)
  clear; intros.
  destruct s'; unfold inv in H; unfold cmemOf, cspOf, cvfOf in * .
  destruct H as [w [v0 [VFE [l1 [l2 [Hp L]]]]]].
  vcSteps. branchStepX.

  (* true branch *)
  (* renaming things to make example look nice *)
  rename v into vf; rename v0 into v1; 
    rename w into v2; rename vs into l0.

  sarewriteE in Hp;
  rename x1 into l1'; rename x0 into v; rename x2 into v1';
  subst l1.
  vcSteps.

  econstructor; econstructor; intuition.

  exists l1'. exists (v::l2). split.
  unfold llist; fold llist. ssimpl. searchMatch.
  clear - H. simpl in *. autorewrite with list in H. assumption.

  (* false branch *)
  subst v0. ssimplg in Hp.
  sarewriteE in Hp.
  subst x. simpl in H. subst x0. assumption.

  (* undefined case *)
  subst v0. sarewriteE in Hp. assumption.

Qed.
