(******************************************************************

   Separation logic toolkit

   Andrew McCreight
   HASP Project
   Portland State University

   http://hasp.cs.pdx.edu/

   Copyright 2009 Portland State University

   This file is part of the separation logic toolkit.

   The separation logic toolkit is free software: you can redistribute
   it and/or modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation, either
   version 3 of the License, or (at your option) any later version.

   The separation logic toolkit is distributed in the hope that it
   will be useful, but WITHOUT ANY WARRANTY; without even the implied
   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with the separation logic toolkit.  If not, see
   <http://www.gnu.org/licenses/>.

*******************************************************************)


Require Import SepCore.
Require Import SepMach.
Require Import SepSimpl.
Require Import SepRearr.
Require Import SepMatch.
Require Import SepSetoid.
Require Import ListAddrSet.
Require Import ZArith.


Set Implicit Arguments.


Open Local Scope word_scope.


Ltac implr :=
  match goal with
    | Hp : (_ ** _) ?m |- (_ ** _) ?m =>
      apply (sconjImplR Hp); clear m Hp; intros m Hp
  end.


(* a new and slightly odd separation logic predicate *)

Definition guard (P:Prop) (A:mpred) : mpred :=
  fun m => (P -> A m) /\ (~P -> emp m).

Lemma guardTrueEqv : forall (P:Prop),
  P -> forall A,
    miff (guard P A) A.
Proof.
  unfold guard; intros P Pok A m; constructor; intros M.
  destruct M; auto.
  constructor; auto; intro; elf; auto.
Qed.
  
Lemma guardFalseEqv : forall (P:Prop),
  ~P -> forall A,
    miff (guard P A) emp.
Proof.
  unfold guard; intros P NP A m; constructor; intros M.
  destruct M; auto.
  constructor; auto; intro; elf; auto.
Qed.


(* add this to sle/sli? *)
Lemma stackFrameFour : forall sp,
  miff (stackFrame sp 4) (sp |->-).
Proof.
  clear; intros sp m; unfold stackFrame.
  replace 4 with (0 + 4)%word by ring.
  rewrite zeroWordSetSplitEnd.
  2:replace (0 + 4)%word with 4 by ring; apply wordLeInj'; omega.
  replace (wordAlign 0) with 0 by (apply wordInj; compute; auto).
  setsimpl.
  assert (I : WS.In 0 (WS.add 0 WS.empty)).
  setsimpl; auto.
  constructor; intro Hp.
  
  hyr Hp (fasE I Hp); cbv beta in Hp; apply empER; ssplit.
  replace (WS.remove 0 (WS.add 0 WS.empty)) with WS.empty in Hp.
  apply (fasEmptyE Hp).
  clear; apply WSE.set_ext; intros x; setsimpl; constructor; intros Ix; tauto.
  
  apply fasI.
  replace (sp + 0)%addr with (sp : addr).
  apply fasEmptyIR; auto.
  clear; simpl; unfold blockAddr; f_equal; ring.
  setsimpl; auto.

Qed.


(* precisely relating sets of addresses and heaps *)

(* (fasDom s) m iff the domain of m is s *)
Definition fasDom s := forallsa x in s, (x:addr) |->- .

Lemma fasDomI : forall x s,
  AS.In x s ->
  mimpl ((x:addr) |->- ** fasDom (AS.remove x s)) (fasDom s).
Proof.
  unfold fasDom; intros x s Ix m Hp.
  replace s with (AS.add x (AS.remove x s)) by setEqSolver.
  apply AFAS.fasI; auto.  rewrite ASE.remove_eqv; intros [X _]; auto.
Qed.

Lemma fasDomE : forall x s,
  AS.In x s ->
  mimpl (fasDom s) ((x:addr) |->- ** fasDom (AS.remove x s)).
Proof.
  unfold fasDom; intros x s Ix m Hp.
  hyr Hp (fasaE Ix Hp); cbv beta in Hp; sle Hp; sli; eauto.
Qed.


Lemma fasDomEmptyI : mimpl emp (fasDom AS.empty).
Proof.
  unfold fasDom; intros m Hp; apply (AFAS.fasEmptyI Hp).
Qed.


Lemma fasDomEmptyE : mimpl (fasDom AS.empty) emp.
Proof.
  unfold fasDom; intros m Hp; apply (AFAS.fasEmptyE Hp).
Qed.


Lemma fasDomEmptyEqv : miff (fasDom AS.empty) emp.
Proof.
  intros m; constructor; intros Hp. apply (fasDomEmptyE Hp).
  apply fasDomEmptyI; auto.
Qed.
Hint Rewrite fasDomEmptyEqv : SLogI.
(*Hint Rewrite <- fasDomEmptyEqv : SLogE.*)


Lemma fasDomDisj : forall S S' m,
  (fasDom S ** fasDom S') m ->
  ASE.Disj S S'.
Proof.
  intros S S' m Hp; apply ASE.DisjI; intros x Ix Ix'.
  srewrite (fasDomE Ix) in Hp; srewrite (fasDomE Ix') in Hp.
  sle Hp; contContSolver Hp.
Qed.


Lemma fasDomDisjL : forall S S' A m,
  (fasDom S ** fasDom S' ** A) m ->
  ASE.Disj S S'.
Proof.
  intros S S' A m Hp; scar Hp; inversion Hp.
  apply (fasDomDisj H0).
Qed.
Hint Resolve fasDomDisjL fasDomDisj.


Lemma fasDomMerge : forall s s',
  mimpl (fasDom s ** fasDom s') (fasDom (s |_| s')).
Proof.
  clear.  unfold fasDom; intros s s' m Hp.
  apply AFAS.fasDisjMerge; [apply (fasDomDisj Hp) | auto].
Qed.


(* Loosely relating sets of addresses and heaps.

   (fasSubDom s) m  iff(?)  s is a subset of the domain of m
*)


Definition fasSubDom (S:AS.t) := forallsa x in S, (x:addr) |->- ** top.


Lemma fasSubDomE : forall x S,
  AS.In x S ->
  mimpl (fasSubDom S) ((x:addr) |->- ** fasSubDom (AS.remove x S) ** top).
Proof.
  unfold fasSubDom; intros x s Ix m Hp.
  hyr Hp (fasaE Ix Hp); cbv beta in Hp; sle Hp; sli; eauto.
Qed.


Lemma fasSubDomDisj : forall S S' m,
  (fasSubDom S ** fasSubDom S') m ->
  ASE.Disj S S'.
Proof.
  intros S S' m Hp; apply ASE.DisjI; intros x Ix Ix'.
  srewrite (fasSubDomE Ix) in Hp; srewrite (fasSubDomE Ix') in Hp.
  sle Hp; contContSolver Hp.
Qed.


Lemma fasSubDomDisjL : forall S S' A m,
  (fasSubDom S ** fasSubDom S' ** A) m ->
  ASE.Disj S S'.
Proof.
  intros S S' A m Hp; scar Hp; inversion Hp.
  apply (fasSubDomDisj H0).
Qed.


Lemma fasMergeBySubDom : forall (P : addr -> mpred),
  (forall x m, P x m -> (x |->- ** top) m) -> forall s s',
  mimpl (AFAS.fas P s ** AFAS.fas P s') (AFAS.fas P (s |_| s')).
Proof.
  intros P Imp s s' m Hp.
  apply AFAS.fasDisjMerge; auto.
  assert ((fasSubDom s ** fasSubDom s') m) as Hp'.
  ssplit; unfold fasSubDom; apply (AFAS.fasImpl Hp); auto.
  apply (fasSubDomDisj Hp').
Qed.


Lemma fasSubDomMerge : forall s s',
  mimpl (fasSubDom s ** fasSubDom s') (fasSubDom (s |_| s')).
Proof.
  unfold fasSubDom; intros s s' m Hp; apply fasMergeBySubDom; auto.
Qed.


Lemma fasDomSubDom : forall s,
  mimpl (fasDom s) (fasSubDom s).
Proof.
  unfold fasDom, fasSubDom; intros s m Hp.
  apply (AFAS.fasImpl Hp); clear m Hp; intros x m Hp.
  hyr Hp (empIR Hp); ssplit.
Qed.

