(******************************************************************

   Separation logic toolkit

   Andrew McCreight
   HASP Project
   Portland State University

   http://hasp.cs.pdx.edu/

   Copyright 2009 Portland State University

   This file is part of the separation logic toolkit.

   The separation logic toolkit is free software: you can redistribute
   it and/or modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation, either
   version 3 of the License, or (at your option) any later version.

   The separation logic toolkit is distributed in the hope that it
   will be useful, but WITHOUT ANY WARRANTY; without even the implied
   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with the separation logic toolkit.  If not, see
   <http://www.gnu.org/licenses/>.

*******************************************************************)


Require Export VFileSig.
Require Import Memory.
Require Export Val.
Require Import FnMap.
Require Import NatUtil.
Require Import OptUtil.


Set Implicit Arguments.


Lemma eq_var_dec : forall (v v':var),
  {v = v'} + {v <> v'}.
Proof.
  intros [x] [x']; destruct (eq_nat_dec x x'); auto.
  right; intro X; injector X; auto.
Defined.


Module VF : VFileSig.

  Inductive vfile0 : Type := vfilei : list var -> map var val -> vfile0.
  Definition vfile := vfile0.

  (* operations *)

  Definition dom (vf:vfile) := let (d, _) := vf in d.

  Definition mapOf (vf:vfile) := let (_, m) := vf in m.

  Definition base vars := vfilei vars (const_map _ Vundef).
  
  Definition update vf x v :=
    if In_dec eq_var_dec x (dom vf) then
      Some (vfilei (dom vf) (map_set eq_var_dec (mapOf vf) x v))
    else
      None.

  Definition get vf x :=
    if In_dec eq_var_dec x (dom vf) then
      Some (map_get (mapOf vf) x)
    else
      None.


  (* domain *)

  Lemma baseDomEq : forall vars,
    dom (base vars) = vars.
  Proof.
    auto.
  Qed.

  Lemma updateDomEq : forall vf x v vf',
    update vf x v = Some vf' ->
    dom vf = dom vf'.
  Proof.
    unfold update; intros vf x v vf' E.
    destruct (In_dec eq_var_dec x (dom vf)) as [_ | _].
    injector E; auto.
    discriminate E.
  Qed.

  Lemma updateDomE : forall vf x,
    In x (dom vf) -> forall v,
      exists vf', update vf x v = Some vf'.
  Proof.
    unfold update; intros [d vv] x; simpl; intros I v.
    destruct (In_dec eq_var_dec x d) as [_ | NI]; [econstructor | elf]; auto.
  Qed.

  Lemma getDomE : forall vf x,
    In x (dom vf) ->
    exists v, get vf x = Some v.
  Proof.
    unfold get; intros [d vv]; simpl; intros x I.
    destruct (In_dec eq_var_dec x d) as [_ |NI]; [econstructor | elf]; auto.
  Qed.


  (* basic operation properties *)
  
  Lemma updateGetEq : forall vf vf' x v,
    update vf x v = Some vf' ->
      get vf' x = Some v.
  Proof.
    unfold update, get; intros [d vv] [d' vv'] x v; simpl; intros E.
    destruct (In_dec eq_var_dec x d) as [I | _].
    2:discriminate E.
    injector E.
    destruct (In_dec eq_var_dec x d') as [_ | NI].
    rewrite map_setEq; auto.
    elf; auto.
  Qed.

  Lemma updateGetNeq : forall vf vf' x v,
    update vf x v = Some vf' -> forall y,
      x <> y ->
      get vf' y = get vf y.
  Proof.
    unfold update, get; intros [d vv] [d' vv'] x v; simpl; intros E y NE.
    destruct (In_dec eq_var_dec x d) as [I | _].
    2:discriminate E.
    injector E.
    destruct (In_dec eq_var_dec y d') as [_ | NI]; auto.
    rewrite map_setNeq; auto.
  Qed.


End VF.


Export VF.


Implicit Arguments updateDomEq [vf x v vf'].
Implicit Arguments updateDomE [vf x].
Implicit Arguments getDomE [vf x].
Implicit Arguments updateGetEq [vf vf' x v].
Implicit Arguments updateGetNeq [vf vf' x v].



(* update a bunch of variables at once *)
Fixpoint vfUpdateVals (vf:vfile) (xx:list var) {struct xx} : 
  list val -> option vfile :=
  fun vv =>
    match xx, vv with
      | nil, nil => Some vf
      | x::xx', v::vv' =>
	do vf' <-- vfUpdateVals vf xx' vv';;;
        VF.update vf' x v
      | _, _ => None
    end.


Lemma vfUpdateValsDomEq : forall argVars args vf vf',
  vfUpdateVals vf argVars args = Some vf' -> 
  dom vf' = dom vf.
Proof.
  induction argVars; destruct args; try (discriminate 1); simpl;
    intros vf vf' E.
  injection E; intro; subst; auto.
  assert (E' : exists vf'', vfUpdateVals vf argVars args = Some vf'').
  destruct (vfUpdateVals vf argVars args);
    [econstructor; auto | discriminate E].
  destruct E' as [vf'' E']; rewrite E' in E.
  rewrite <- (updateDomEq E).
  apply (IHargVars _ _ _ E'); auto.
Qed.
Implicit Arguments vfUpdateValsDomEq [argVars args vf vf'].


Lemma argInitOk' : forall args argVars,
  length args = length argVars -> forall ll,
    (forall x, In x argVars -> In x ll) ->
    exists vf', 
      vfUpdateVals (base ll) argVars args = Some vf'.
Proof.
  induction args; destruct argVars; intros E ll Sub; try (discriminate E);
    simpl.
  econstructor; reflexivity.
  simpl in E; injection E; clear E; intro E.
  destruct (IHargs _ E ll) as [vf' E']; clear E IHargs.
  intros z Iz; apply Sub; right; auto.
  rewrite E'.
  apply updateDomE.
  rewrite (vfUpdateValsDomEq E').
  rewrite baseDomEq; auto; apply Sub; simpl; auto.
Qed.


Lemma argInitOk : forall args vals,
  length args = length vals -> forall locs,
    exists vf, vfUpdateVals (base (args ++ locs)) args vals = Some vf.
Proof.
  intros args vals E locs; eapply argInitOk'; eauto.
  clear E vals; induction args; simpl; intros; [elf; auto | ].
  destruct H; auto.
Qed.
Implicit Arguments argInitOk [args vals].





(* a nicer way to reason about vfiles *)

Fixpoint listMapGet (A:Set) (l:list (prod var A)) (x:var) {struct l}
  : option A :=
  match l with
    | nil => None
    | (y, v)::l' =>
      if eq_var_dec x y then
        Some v
      else
        listMapGet l' x
  end.

Inductive vfEqv (base : list var) (vals : list (prod var val)) (vf : vfile)
  : Prop :=
  vfEqvi :
    (forall x, In x base -> In x (dom vf)) ->
    (forall x v, listMapGet vals x = Some v -> get vf x = Some v) ->
    vfEqv base vals vf.
      (* also require In x base? *)

Fixpoint zip A B (l1:list A) (l2:list B) {struct l1} : list (prod A B) :=
  match l1, l2 with
    | x::l1', v::l2' => (x,v)::zip l1' l2'
    | _, _ => nil
  end.

Fixpoint inMapListDom A (x:A) (ll:list (prod A val)) {struct ll} : Prop :=
  match ll with
    | nil => False
    | (y,v)::ll' => x = y \/ inMapListDom x ll'
  end.

Definition beq_var (x y:var) :=
  match x, y with
    | Var x, Var y => beq_nat x y
  end.

Fixpoint mapListFind (ll:list (prod var val)) (x:var) {struct ll} : val :=
  match ll with
    | nil => Vundef
    | (y,v)::ll' =>
      if beq_var x y then
	v
      else
	mapListFind ll' x
  end.



(* properties of vfEqv *)

Lemma vfEqvI : forall bvars ivars ivals vf,
  vfUpdateVals (base bvars) ivars ivals = Some vf ->
  vfEqv bvars (zip ivars ivals) vf.
Proof.
  induction ivars; destruct ivals; simpl; intros; try (discriminate H).

  injector H; constructor.
  intros x Ix; rewrite baseDomEq; auto.
  discriminate 1.

  assert (E : exists vf', vfUpdateVals (base bvars) ivars ivals = Some vf').
  destruct (vfUpdateVals (base bvars) ivars ivals); [econstructor | elf]; auto.
  discriminate H.

  destruct E as [vf' E]; rewrite E in H; hyr IHivars (IHivars _ _ E).
  constructor.
  intros x Ix.
  rewrite <- (updateDomEq H); rewrite (vfUpdateValsDomEq E); 
    rewrite baseDomEq; auto.
  simpl; intros x v' II.
  destruct (eq_var_dec x a) as [E' | NE].
  subst; injector II; apply (updateGetEq H).
  destruct IHivars as [_ X]; hyr X (X _ _ II).
  rewrite (updateGetNeq H); auto.
Qed.
Implicit Arguments vfEqvI [bvars ivars ivals vf].


Lemma vfEqvGetEq : forall b ups vf,
  vfEqv b ups vf -> forall x,
    inMapListDom x ups ->
    get vf x = Some (mapListFind ups x).
Proof.
  intros b ups vf [_ Y] x Ix; hyr Y (Y x).
  generalize dependent ups.
  induction ups.
  intros; elf; auto.

  destruct a as [y v]; simpl.
  destruct (eq_var_dec x y) as [E | NE].

  subst y; intros _ E; unfold beq_var; destruct x; rewrite beq_true; auto.

  unfold beq_var; destruct x; destruct y; rewrite beq_false; auto.
  intros [E | Ix] X; [elf | ]; eauto.
Qed.


Lemma vfUpdateOk : forall bvars ups vf,
  vfEqv bvars ups vf -> forall x v,
    In x bvars ->
      exists vf', 
	(forall vf0, vf = vf0 -> update vf0 x v = Some vf') /\
	vfEqv bvars ((x,v)::ups) vf'.
Proof.
  intros bvars ups vf VFE x v Ix.

  assert (VV : In x (dom vf)).
  destruct VFE as [X _]; auto.

  destruct (updateDomE VV v) as [vf' E]; exists vf'; constructor.
  intros; subst; auto.

  destruct VFE as [VE1 VE2]; constructor.
  intros y Iy; rewrite <- (updateDomEq E); auto.
  simpl; intros y v' Iy; destruct (eq_var_dec x y) as [E' | NE].
  subst; simpl; destruct (eq_var_dec y y) as [_ | NE]; [injector Iy |elf; auto].
  rewrite (updateGetEq E); auto.

  destruct (eq_var_dec y x) as [E' | _].
  elf; auto.
  rewrite (updateGetNeq E); auto.
Qed.


Lemma vfEqvWeaken : forall bv up vf,
  vfEqv bv up vf -> forall bv' up',
    (forall x, In x bv' -> In x bv) ->
    (forall x v, listMapGet up' x = Some v -> listMapGet up x = Some v) ->
    vfEqv bv' up' vf.
Proof.
  intros bv up vf [VF VF'] bv' up' VFi VF'i; constructor; eauto.
Qed.


(** a bunch of infrastructure to reason about vfEqv,
    culminating in vfEqvSolver  *)


(* drop any pairs where the key appears twice *)
Fixpoint dropDups' (l:list (prod var val)) : list var -> list (prod var val) :=
  fun dup =>
    match l with
      | nil => nil
      | (x, v)::l' =>
        if In_dec eq_var_dec x dup then
          dropDups' l' dup
        else
          (x, v):: dropDups' l' (x::dup)
    end.
Definition dropDups l := dropDups' l nil.

(* dropDups' doesn't lose values *)
Lemma dropDups'Ok1 : forall l x v,
  listMapGet l x = Some v -> forall dup,
  ~In x dup ->
  listMapGet (dropDups' l dup) x = Some v.
Proof.
  induction l; simpl; intros x v E dup NI.

  discriminate E.

  destruct a as [x' v'].
  destruct (eq_var_dec x x') as [E' | NE].
  subst; injector E.
  destruct (In_dec eq_var_dec x' dup) as [I | _].
  elf; auto.
  simpl; destruct (eq_var_dec x' x') as [_ | NE]; auto; elf; auto.

  hyr IHl (IHl _ _ E).
  destruct (In_dec eq_var_dec x' dup) as [I | NI'].
  auto.
  simpl.
  destruct (eq_var_dec x x') as [E' | _].
  elf; auto.
  apply IHl; simpl; intuition.
Qed.


(* dropDups' doesn't add values *)
Lemma dropDups'Ok2 : forall l dup x v,
  listMapGet (dropDups' l dup) x = Some v -> 
  ~In x dup ->
  listMapGet l x = Some v.
Proof.
  induction l; simpl; intros dup x v E NI.

  discriminate E.

  destruct a as [x' v'].
  destruct (eq_var_dec x x') as [E' | NE].
  subst; injector E.
  destruct (In_dec eq_var_dec x' dup) as [I | _].
  elf; auto.
  simpl in H; destruct (eq_var_dec x' x') as [_ | NE]; auto; elf; auto.

  destruct (In_dec eq_var_dec x' dup) as [I | NI'].
  hyr IHl (IHl _ _ _ E); auto.
  simpl in E.
  destruct (eq_var_dec x x') as [E' | _].
  elf; auto.
  apply (IHl _ _ _ E); simpl; intuition.
Qed.


(* dropping duplicates results in an equivalent list *)
Lemma dropDupsOk : forall l x v,
  listMapGet l x = Some v <-> 
  listMapGet (dropDups l) x = Some v.
Proof.
  unfold dropDups; intros l x v; constructor; intro E.
  apply dropDups'Ok1; auto.
  eapply dropDups'Ok2; eauto.
Qed.


Lemma listMapGetEq : forall l x (v:val),
  listMapGet l x = Some v ->
  In (x, v) l.
Proof.
  induction l; simpl; intros x v E.
  discriminate E.
  destruct a as [x' v'].
  destruct (eq_var_dec x x') as [E' | NE].
  injector E; auto.
  right; auto.
Qed.


Lemma inDupsNotEqv : forall l dups x v,
  In (x, v) (dropDups' l dups) ->
  In x dups ->
  False.
Proof.
  induction l; simpl; intros dups x v E I.
  auto.
  destruct a as [x' v'].
  destruct (In_dec eq_var_dec x' dups) as [I' | NI].
  eapply IHl; eauto.
  destruct (eq_var_dec x x') as [E' | NE];
    simpl in E; destruct E; subst; auto.
  injection H; auto.
  eapply IHl.
  apply H.
  simpl; auto.
Qed.

Lemma listMapGetDupEqv' : forall l dups x v,
  In (x, v) (dropDups' l dups) ->
  ~In x dups ->
  listMapGet (dropDups' l dups) x = Some v.
Proof.
  induction l; simpl; intros dups x v E NI.
  elf; auto.
  destruct a as [x' v'].
  destruct (In_dec eq_var_dec x' dups) as [I | NI'].
  apply IHl; auto.

  simpl in E |- * .
  destruct (eq_var_dec x x') as [E' | NE];
    subst; destruct E as [E1 | E].
  injector E1; auto.
  elf; eapply inDupsNotEqv; eauto; simpl; auto.
  injector E1; elf; auto.
  apply IHl; auto; simpl; intuition.
Qed.

Lemma listMapGetDupEqv : forall l x v,
  listMapGet l x = Some v <->
  In (x, v) (dropDups l).
Proof.
  intros l x v; constructor; intro E.

  rewrite dropDupsOk in E; apply listMapGetEq in E; auto.

  rewrite dropDupsOk; unfold dropDups in *;
    eapply listMapGetDupEqv'; eauto.
Qed.


Lemma vfEqvDropDups : forall b vv vf,
  vfEqv b vv vf ->
  vfEqv b (dropDups vv) vf.
Proof.
  intros b vv vf VFE.
  apply (vfEqvWeaken VFE).
  auto.
  intros x v E; rewrite <- dropDupsOk in E; assumption.
Qed.


Ltac vfEqvSolver0 :=
  match goal with
    | VE : vfEqv _ _ _ |- vfEqv _ _ _ =>
      apply (vfEqvWeaken VE);
        [auto; simpl; tauto |
          let x := fresh "x" in
          let v := fresh "v" in
          let E := fresh "E" in
            intros x v E;
              rewrite listMapGetDupEqv in E;
                rewrite listMapGetDupEqv; simpl in E |- * ;
                  intuition eassumption]
  end.



(* improved vfEqvSolver *)

Lemma pairInj : forall A B (a a':A) (b b':B),
  (a, b) = (a', b') ->
  a = a' /\ b = b'.
Proof.
  intros A B a a' b b' E; injector E; auto.
Qed.

Ltac vfEqvOrSplit Iz :=
  match type of Iz with
    | False => elf; exact Iz
    | _ \/ _ => destruct Iz as [Iz | Iz]; [ | vfEqvOrSplit Iz]
  end.

Ltac vfEqvSolverGoalMatch :=
  match goal with
    | |- (?x, _) = (?x, _) \/ _ => left; f_equal; eauto
    | |- _ \/ _ => right; vfEqvSolverGoalMatch
    | |- False => idtac (* var doesn't exist! *)
  end.


(* machinery for simplifying dropDups *)

Lemma In_decEq : forall x l A (e e':A),
  (if In_dec eq_var_dec x (x::l) then e else e') = e.
Proof.
  intros x l A e e'.
  simpl; destruct (eq_var_dec x x); auto; elf; auto.
Qed.

Lemma In_decNeq : forall x y l A (e e':A),
  x <> y ->
  (if In_dec eq_var_dec x (y::l) then e else e') =
  (if In_dec eq_var_dec x l then e else e').
Proof.
  intros x y l A e e' NE.
  simpl; destruct (eq_var_dec y x).
  elf; auto.
  destruct (In_dec eq_var_dec x l); auto.
Qed.

Lemma In_decNil : forall x A (e e':A),
  (if In_dec eq_var_dec x nil then e else e') = e'.
Proof.
  intros x A e e'; reflexivity.
Qed.


Hint Rewrite In_decEq In_decNil : In_decSimpl.
Hint Rewrite In_decNeq using (clear; discriminate) : In_decSimpl.

Lemma dropDups'Step : forall x v l dup,
  dropDups' ((x, v)::l) dup =
  (if In_dec eq_var_dec x dup then
     dropDups' l dup
   else
     (x, v):: dropDups' l (x::dup)).
Proof.
  reflexivity.
Qed.


Lemma dropDups'StepNil : forall l,
  dropDups' nil l = nil.
Proof.
  reflexivity.
Qed.

Ltac dropDupsSimpl VFE :=
  unfold dropDups in VFE;
    repeat (rewrite dropDups'Step in VFE; 
      autorewrite with In_decSimpl in VFE);
    rewrite dropDups'StepNil in VFE.

Ltac dropDupsSimplG :=
  unfold dropDups;
    repeat (rewrite dropDups'Step; autorewrite with In_decSimpl);
    rewrite dropDups'StepNil.


Ltac dropVfeDups VFE :=
  apply vfEqvDropDups in VFE; dropDupsSimpl VFE.


Ltac vfEqvSolver :=
  match goal with
    | VFE : vfEqv _ _ _ |- vfEqv _ _ _ =>
      apply (vfEqvWeaken VFE); clear VFE; [auto; simpl; tauto |
        let x := fresh "x" in
        let v := fresh "v" in
        let Ix := fresh "Ix" in
          intros x v Ix;
          rewrite listMapGetDupEqv in Ix |- * ;
          dropDupsSimpl Ix; dropDupsSimplG;
          cbv iota beta delta [In] in Ix |- * ;
          vfEqvOrSplit Ix;
          apply pairInj in Ix;
          let E1 := fresh "E" in
          let E2 := fresh "E" in
            destruct Ix as [E1 E2]; subst x; subst v;
              vfEqvSolverGoalMatch]
  end.


Hint Extern 3 (vfEqv ?X ?Y ?Z) => vfEqvSolver.