(******************************************************************

   Separation logic toolkit

   Andrew McCreight
   HASP Project
   Portland State University

   http://hasp.cs.pdx.edu/

   Copyright 2009 Portland State University

   This file is part of the separation logic toolkit.

   The separation logic toolkit is free software: you can redistribute
   it and/or modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation, either
   version 3 of the License, or (at your option) any later version.

   The separation logic toolkit is distributed in the hope that it
   will be useful, but WITHOUT ANY WARRANTY; without even the implied
   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with the separation logic toolkit.  If not, see
   <http://www.gnu.org/licenses/>.

*******************************************************************)


Require Import CoreLog.
Require Import ZArith.
Require Export ZUtil.
Require Export NatUtil.
Require Export Ring.


Set Implicit Arguments.





(* explicitly declaring this should make things easier *)
Definition wordBound := 4294967296%Z.
  (* two_power_nat 32 *)
Definition indexBound := 1073741824%Z.
Definition okWordZ n := (0 <= n < wordBound)%Z.
Definition okWordNat n := (0 <= Z_of_nat n < wordBound)%Z.
Record word : Set := wordi {wval : Z; wok : okWordZ wval}.
  (* following Leroy's CompCert we use a record instead of a sig
     to avoid problems with the injection tactic. *)


Delimit Scope word_scope with word.


Open Local Scope word_scope.


Lemma wordBoundPos :
  (wordBound > 0)%Z.
Proof.
  unfold wordBound; omega.
Qed.
Hint Resolve wordBoundPos : Word.


Lemma zeroWord : okWordZ 0%Z.
Proof.
  constructor; unfold wordBound; omega.
Qed.

Lemma oneWord : okWordZ 1%Z.
Proof.
  constructor; unfold wordBound; omega.
Qed.

Lemma twoWord : okWordZ 2%Z.
Proof.
  constructor; unfold wordBound; omega.
Qed.

Lemma threeWord : okWordZ 3%Z.
Proof.
  constructor; unfold wordBound; omega.
Qed.

Lemma fourWord : okWordZ 4%Z.
Proof.
  constructor; unfold wordBound; omega.
Qed.

Hint Resolve zeroWord oneWord twoWord threeWord fourWord : Word.

(* operations *)

Hint Resolve Z_mod_lt : Word.

Ltac wauto := auto with * .

Lemma boundModOk : forall a,
  (0 <= a mod wordBound < wordBound)%Z.
Proof.
  intros; apply Z_mod_lt; wauto.
Qed.


Lemma Z_to_wordOk : forall x,
  okWordZ (x mod wordBound)%Z.
Proof.
  intro x; unfold okWordZ; apply boundModOk.
Qed.

Definition Z_to_word (x:Z) : word := wordi (Z_to_wordOk x).


Lemma nat_to_wordOk : forall n,
  okWordZ ((Z_of_nat n) mod wordBound)%Z.
Proof.
  intro n; unfold okWordZ; apply boundModOk.
Qed.

Definition nat_to_word (n:nat) : word := wordi (nat_to_wordOk n).

Coercion nat_to_word : nat >-> word.

Notation "0" := (0:word) : word_scope.
Notation "1" := (1:word) : word_scope.
Notation "2" := (2:word) : word_scope.
Notation "3" := (3:word) : word_scope.
Notation "4" := (4:word) : word_scope.
Definition windexBound := (Z_to_word indexBound).

Definition bool_to_word (b:bool) : word :=
  if b then 1 else 0.

Coercion bool_to_word : bool >-> word.


Definition wordOp (f:Z -> Z -> Z) (w w':word) : word :=
  wordi (boundModOk (f (wval w) (wval w'))%Z).

Definition wordAdd := wordOp (fun x y => x + y)%Z.
Definition wordSub := wordOp (fun x y => x - y)%Z.
Definition wordMult := wordOp (fun x y => x * y)%Z.
Definition wordDiv := wordOp (fun x y => x / y)%Z.
Definition wordMod := wordOp (fun x y => x mod y)%Z.
Definition wordNeg x := wordSub 0 x.

Infix "+" := wordAdd : word_scope.
Infix "-" := wordSub : word_scope.
Infix "*" := wordMult : word_scope.
Infix "/" := wordDiv : word_scope.
Infix "mod" := wordMod : word_scope.
Notation "- x" := (wordNeg x) : word_scope.

Definition wordP (f:Z -> Z -> Prop) (w w':word) : Prop :=
  f (wval w) (wval w').

Definition wordLt := wordP (fun x y => x < y)%Z.
Definition wordLte := wordP (fun x y => x <= y)%Z.

Infix "<" := wordLt : word_scope.
Infix "<=" := wordLte : word_scope.

Definition word_to_nat (w:word) : nat := Zabs_nat (wval w).
Definition word_to_Z (w:word) : Z := wval w.


Lemma wordModEq : forall a,
  okWordZ a -> 
  (a mod wordBound = a)%Z.
Proof.
  intros; unfold wordBound; apply ltModEq; wauto.
Qed.


Lemma zeroModEq : (0 mod wordBound = 0)%Z.
Proof.
  rewrite wordModEq; wauto.
Qed.

Lemma oneModEq : (1 mod wordBound = 1)%Z.
Proof.
  rewrite wordModEq; wauto.
Qed.

Lemma twoModEq : (2 mod wordBound = 2)%Z.
Proof.
  rewrite wordModEq; wauto.
Qed.

Lemma threeModEq : (3 mod wordBound = 3)%Z.
Proof.
  rewrite wordModEq; wauto.
Qed.

Lemma fourModEq : (4 mod wordBound = 4)%Z.
Proof.
  rewrite wordModEq; wauto.
Qed.

Ltac wunfold :=
  unfold wordNeg, wordLt, wordLte, wordP,
    wordSub, wordAdd, wordMult, wordDiv,
    wordOp, okWordZ, okWordNat, word_to_nat in * ; simpl in * ;
      try rewrite zeroModEq in * ; try rewrite oneModEq in * ;
        try rewrite twoModEq in * ; try rewrite threeModEq in * ;
          try rewrite fourModEq in * .

Lemma wordNeInj : forall (x y:word),
  x <> y ->
  wval x <> wval y.
Proof.
  intros [x H] [y H'] NE E; simpl in E; subst y; apply NE;
    rewrite (proof_irrel H H'); auto.
Qed.


Lemma wordNeInj' : forall (x y:word),
  wval x <> wval y ->
  x <> y.
Proof.
  intros [x X] [y Y] NE E; simpl in * ; injection E; auto.
Qed.


Lemma eq_word_dec : forall (w w':word), {w = w'} + {w <> w'}.
Proof.
  intros [x H] [y H'].
  destruct (eq_Z_dec x y); [left | right].
  subst y; rewrite (proof_irrel H H'); auto.
  intro E; injection E; auto.
Qed.


Definition beq_word (w w':word) :=
  match (wval w ?= wval w')%Z with
    | Eq => true
    | _ => false
  end.

Lemma beq_wordTrue : forall w,
  beq_word w w = true.
Proof.
  unfold beq_word; intros [w wOk]; rewrite Zcompare_refl; auto.
Qed.

Lemma beq_wordFalse : forall w w',
  w <> w' ->
  beq_word w w' = false.
Proof.
  unfold beq_word; intros [w wOk] [w' w'Ok] NE; hyr NE (wordNeInj NE); simpl in * .
  assert (E := Zcompare_Eq_eq w w').
  destruct (w ?= w')%Z; auto; elf; auto.
Qed.

Lemma lt_word_dec : forall (m n:word),
  {m < n} + {~(m < n)}.
Proof.
  intros [x H] [y H']; wunfold; apply zlt_dec.
Qed.


Lemma lte_word_dec : forall (m n:word),
  {m <= n} + {n < m}.
Proof.
  intros m n.
  destruct (eq_word_dec m n) as [E | NE].
  left; subst; wunfold; omega.
  destruct (lt_word_dec m n) as [LT | NLT].
  left; wunfold; omega.
  hyr NE (wordNeInj NE); right; wunfold; omega.
Qed.


Lemma word_to_natLtI : forall x y,
  x < y ->
  (word_to_nat x < word_to_nat y)%nat.
Proof.
  intros [x H] [y H'] LT; wunfold.
  assert (D : (Zabs_nat x < Zabs_nat y \/ Zabs_nat x >= Zabs_nat y)%nat); 
    [omega | destruct D as [D | D]]; auto.
  hyr D (inj_ge _ _ D).
  rewrite Z_to_nat_to_Z in D; [ | omega].
  rewrite Z_to_nat_to_Z in D; omega.
Qed.


Lemma word_to_natLtE : forall x y,
  (word_to_nat x < word_to_nat y)%nat ->
  x < y.
Proof.
  intros [x H] [y H'] LT; wunfold.
  assert (D : (x < y \/ x = y \/ y < x)%Z);
    [omega | destruct D as [D | [D | D]]]; auto.
  subst y; omega.
  assert (X := word_to_natLtI (x := wordi H') (y := wordi H)).
  wunfold; hyr X (X D); omega.
Qed.


Definition word_to_bool (w:word) : bool :=
  if eq_word_dec w 0 then false else true.


Lemma okWordNatWord : forall x,
  okWordNat (word_to_nat x).
Proof.
  intros [x H]; wunfold.
  rewrite Z_to_nat_to_Z; auto; omega.
Qed.


Lemma wordInj : forall (x y:word),
  wval x = wval y ->
  x = y.
Proof.
  intros x y E; destruct (eq_word_dec x y); auto; elf.
  destruct x as [x X]; destruct y as [y Y]; wunfold; apply n; clear n; subst.
  rewrite (proof_irrel X Y); auto.
Qed.


Lemma word_to_nat_to_word : forall (i:word),
  nat_to_word (word_to_nat i) = i.
Proof.
  intros i.
  apply wordInj.
  destruct i; wunfold.
  rewrite Z_to_nat_to_Z.
  rewrite wordModEq; auto.
  omega.
Qed.

  
Lemma nat_to_wordSEq : forall n,
  nat_to_word (S n) = nat_to_word n + 1.
Proof.
  intros n; apply wordInj.
  unfold wval, nat_to_word; rewrite inj_S.
  simpl.
  assert (GT : (Z_of_nat n >= 0)%Z).
  induction n.
  simpl; omega.
  rewrite inj_S; omega.

  generalize dependent (Z_of_nat n); clear n; intros.
  assert (wbPos : (wordBound > 0)%Z).
  wauto.
  replace (Zsucc z) with (z + 1)%Z by auto.
  assert (X := Z_div_mod_eq z _ wbPos).
  assert (XY : (z + 1 = (z mod wordBound + 1) + (z / wordBound) * wordBound)%Z).
  replace ((z / wordBound) * wordBound)%Z with (wordBound * (z / wordBound))%Z.
  2:auto with * .
  wauto.
  rewrite XY; clear XY.
  rewrite Z_mod_plus; auto.
Qed.


Lemma nat_to_word_to_nat : forall x,
  okWordNat x ->
  word_to_nat (nat_to_word x) = x.
Proof.
  intros x X; wunfold.
  rewrite wordModEq; auto.
  rewrite nat_to_Z_to_nat; auto.
Qed.



(* there's a ring over words *)


Lemma wordAddSym : forall x y,
  x + y = y + x.
Proof.
  intros [x xOk] [y yOk]; apply wordInj; wunfold; clear xOk yOk.
  replace (x + y)%Z with (y + x)%Z; omega.
Qed.


Lemma wordAddZeroEq : forall x,
  x + 0 = x.
Proof.
  intros [x xOk]; apply wordInj; wunfold.
  replace (x + 0)%Z with x by omega.
  rewrite wordModEq; auto.
Qed.

Lemma wordZeroAddEq : forall x,
  0 + x = x.
Proof.
  intro x. rewrite wordAddSym. rewrite wordAddZeroEq. reflexivity.
Qed.

Lemma wordAddAssoc : forall x y z,
  x + (y + z) = x + y + z.
Proof.
  intros [x xOk] [y yOk] [z zOk]; symmetry; apply wordInj; wunfold.
  transitivity ((x + y + z) mod wordBound mod wordBound)%Z.
  rewrite plusModDistrib; wauto; rewrite modModEqEq; wauto.
  rewrite modModEqEq; wauto.
  replace (x + (y + z) mod wordBound)%Z with
    ((y + z) mod wordBound + x)%Z by omega.
  rewrite plusModDistrib; wauto.
  replace (x + y + z)%Z with (y + z + x)%Z by omega; auto.
Qed.



(* wordAddAssoc *)

Lemma wordMult1 : forall x,
  1 * x = x.
Proof.
  intros [x xOk]; apply wordInj.
  unfold wordMult, wordOp, wval.
  unfold nat_to_word.
  replace (Z_of_nat 1 mod wordBound)%Z with 1%Z.
  replace (1 * x)%Z with x by ring.
  rewrite wordModEq; auto.
  simpl; rewrite oneModEq; auto.
Qed.

Lemma wordMulSym : forall x y,
  x * y = y * x.
Proof.
  intros [x xOk] [y yOk]; apply wordInj; wunfold.
  replace (x * y)%Z with (y * x)%Z by auto with * ; auto.
Qed.


Lemma wordMultAssoc : forall x y z,
  x * (y * z) = x * y * z.
Proof.
  intros [x xOk] [y yOk] [z zOk]; apply wordInj; wunfold.
  transitivity ((x * y * z) mod wordBound mod wordBound)%Z.
  rewrite Zmult_comm.
  rewrite multModDistrib; wauto; rewrite modModEqEq; wauto.
  replace (y * z * x)%Z with (x * y * z)%Z by ring; auto.
  rewrite modModEqEq; wauto.
  rewrite multModDistrib; wauto.
Qed.


Lemma wordDistr : forall x y z,
  (x + y) * z = x * z + y * z.
Proof.
  intros [x xOk] [y yOk] [z zOk]; apply wordInj; wunfold.
  rewrite modPlusEq; wauto.
  replace (x * z + y * z)%Z with ((x + y) * z)%Z by ring.
  rewrite multModDistrib; wauto.
Qed.


Lemma wordSubNegEq : forall x y,
  x - y = x + -y.
Proof.
  intros [x xOk] [y yOk]; apply wordInj; wunfold.
  replace (x - y)%Z with (x + -y)%Z by auto.
  rewrite <- modPlusEq; wauto.
  rewrite (@wordModEq x); wauto.
Qed.


Lemma wordMinusEqZero : forall x,
  x - x = 0.
Proof.
  intros [x xOk]; apply wordInj; wunfold.
  replace (x - x)%Z with 0%Z by omega.
  rewrite wordModEq; wauto.
Qed.


Lemma wordNegEqZero : forall x,
  x + - x = 0.
Proof.
  intros x; rewrite <- wordSubNegEq; apply wordMinusEqZero.
Qed.

Definition wordRingTheory :=
  mk_rt 0 1 wordAdd wordMult wordSub wordNeg (@eq _)
  wordZeroAddEq wordAddSym wordAddAssoc wordMult1 wordMulSym wordMultAssoc
  wordDistr wordSubNegEq wordNegEqZero.

Definition wordEqTest (x y:word) : bool :=
  if eq_word_dec x y then true else false.

Lemma wordEqTestOk : forall x y,
  wordEqTest x y = true -> x = y.
Proof.
  unfold wordEqTest; intros x y E.
  destruct (eq_word_dec x y); auto; discriminate E.
Qed.

Add Ring WordRing : wordRingTheory. (* (decidable wordEqTestOk).*)
  (* for some reason, the decidable thing causes problems *)

(* ring has been defined *)


Lemma wordAddSubAssoc : forall x y z,
  x + y - z = x + (y - z).
Proof.
  intros; ring.
Qed.

Lemma wordSubAddEq : forall x y,
  x - y + y = x.
Proof.
  intros; ring.
Qed.

Lemma wordAddSubEq : forall x y,
  x + y - y = x.
Proof.
  intros; ring.
Qed.


(* below 2 are special cases of this *)

Lemma predSuccEq : forall w,
  (w + 1) - 1 = w.
Proof.
  intros w; ring.
Qed.

Lemma succPredEq : forall w,
  (w - 1) + 1 = w.
Proof.
  intros; ring.
Qed.

Lemma wpredLt : forall w,
  0 <> w ->
  w - 1 < w.
Proof.
  intros [w W] NE; hyr NE (wordNeInj NE); wunfold.
  rewrite wordModEq.
  omega.
  wunfold; omega.
Qed.

(* general induction principle for words *)
Theorem wordInd : forall P : word -> Prop,
  P 0 ->
  (forall w, (forall w', w' < w -> P w') -> 0 <> w -> P w) ->
  forall w, P w.
Proof.
  intros P P0 PS w.
  rewrite <- word_to_nat_to_word.
  assert (wOk : okWordNat (word_to_nat w)); [apply okWordNatWord | ].
  generalize dependent (word_to_nat w); clear w.
  induction n using natInd'; intro nOk.

  auto.

  apply PS.
  intros w' LT.
  hyr H (H (word_to_nat w')).
  rewrite word_to_nat_to_word in H.
  apply H.
  replace n with (word_to_nat (nat_to_word n)).
  apply word_to_natLtI; auto.
  rewrite nat_to_word_to_nat; auto.
  apply okWordNatWord.

  apply wordNeInj'; simpl; rewrite zeroModEq; rewrite wordModEq; [omega | auto].

Qed.

Theorem wordInd1 : forall (P:word -> Prop),
  P 0 ->
  (forall w, P w -> 0 <> w + 1 -> P (w + 1)) -> 
  forall w, P w.
Proof.
  intros.
  induction w using wordInd; auto.
  hyr H0 (H0 (w - 1)).
  rewrite succPredEq in H0; apply H0; clear H0; auto.
  apply H1; apply wpredLt; auto.
Qed.

Theorem wordInd4 : forall (P:word -> Prop),
  P 0 ->
  P 1 ->
  P 2 ->
  P 3 ->
  (forall w, P w -> 4 <= w + 4 -> P (w + 4)) -> 
  forall w, P w.
Proof.
  intros P P0 P1 P2 P3 Px.
  induction w using wordInd; auto.
  destruct (eq_word_dec w 1) as [E | NE].
  subst; auto.
  destruct (eq_word_dec w 2) as [E | NE'].
  subst; auto.
  destruct (eq_word_dec w 3) as [E | NE''].
  subst; auto.
  hyr Px (Px (w - 4)).
  clear P0 P1 P2 P3.
  apply wordNeInj in H0; apply wordNeInj in NE;
    apply wordNeInj in NE'; apply wordNeInj in NE''.
  simpl in * .
  rewrite zeroModEq in H0.
  rewrite oneModEq in NE; rewrite twoModEq in NE'; rewrite threeModEq in NE''.
  rewrite wordSubAddEq in Px; apply Px; clear Px; auto.
  apply H; destruct w as [w wOk]; wunfold.
  rewrite wordModEq; wunfold; omega.
  destruct w; wunfold; omega.
Qed.


Lemma wordLtInj : forall (x y:word),
  x < y ->
  (wval x < wval y)%Z.
Proof.
  intros [x H] [y H'] LT; wunfold; auto.
Qed.


Lemma wordLeInj : forall (x y:word),
  x <= y ->
  (wval x <= wval y)%Z.
Proof.
  intros [x H] [y H'] LT; wunfold; auto.
Qed.


Lemma wordSuccLtE : forall x y,
  x < y + 1 ->
  x <= y.
Proof.
  intros [x xOk] [y yOk] LT; hyr LT (wordLtInj LT); wunfold.
  destruct (eq_Z_dec y (wordBound - 1)).
  subst y.
  replace (wordBound - 1 + 1)%Z with wordBound in LT by auto.
  rewrite Z_mod_same in LT; wauto.
  rewrite wordModEq in LT; wauto.
  constructor; wauto.
Qed.


Lemma wordSuccLtI : forall x y,
  x < y ->
  0 <> y + 1 ->
  x < y + 1.
Proof.
  intros [x xOk] [y yOk] LT NE; hyr LT (wordLtInj LT); hyr NE (wordNeInj NE).
  wunfold.
  assert (E : ((y + 1) mod wordBound = y + 1)%Z).
  rewrite wordModEq; auto.
  unfold okWordZ.
  assert ((y + 1 <> wordBound)%Z).
  intro E; rewrite E in NE; rewrite Z_mod_same in NE; wauto.
  wauto.
  rewrite E; omega.
Qed.


Lemma wordSuccLtIEq : forall w,
  0 <> w + 1 ->
  w < w + 1.
Proof.
  intros [w wOk] NE; hyr NE (wordNeInj NE); wunfold.
  rewrite wordModEq.
  omega.
  assert ((w + 1)%Z <> wordBound).
  intro E; rewrite E in NE; rewrite Z_mod_same in NE; wauto.
  wunfold; omega.
Qed.


Lemma word_to_natEqE : forall w w',
  word_to_nat w = word_to_nat w' ->
  w = w'.
Proof.
  intros w w' E.
  destruct (eq_word_dec w w') as [E' | NE]; auto.
  hyr NE (wordNeInj NE); elf; apply NE; clear NE.
  destruct w as [w W]; destruct w' as [w' W']; wunfold.
  replace w with (Z_of_nat (Zabs_nat w)).
  replace w' with (Z_of_nat (Zabs_nat w')).
  rewrite E; auto.
  rewrite Z_to_nat_to_Z; omega.
  rewrite Z_to_nat_to_Z; omega.
Qed.

Lemma wordSEq : forall w,
  (0:word) <> w + 1 ->
  word_to_nat (w + 1) = S (word_to_nat w).
Proof.
  intros w NE; hyr NE (wordNeInj NE).
  destruct w as [w wOk]; wunfold.
  rewrite wordModEq.
  clear NE.
  transitivity (Zabs_nat (Z_of_nat (Zabs_nat w) + Z_of_nat 1)).
  rewrite Z_to_nat_to_Z; auto; omega.
  rewrite <- inj_plus; rewrite nat_to_Z_to_nat; omega.
  wunfold.
  destruct (eq_Z_dec (w + 1) wordBound) as [E | NE'].
  elf; apply NE; rewrite E; rewrite Z_mod_same; wauto.
  omega.
Qed.


Lemma wsuccLt : forall i i',
  i + 1 <= i' ->
  0 <> i + 1 ->
  i < i'.
Proof.
  intros [i iOk] [i' i'Ok] LE NE; hyr NE (wordNeInj NE); hyr LE (wordLeInj LE).
  wunfold.
  rewrite wordModEq in LE.
  omega.
  constructor.
  omega.
  destruct (eq_Z_dec (i + 1) wordBound) as [E | NE'].
  elf; rewrite E in NE; rewrite Z_mod_same in NE; wauto.
  omega.
Qed.


Lemma wordMultEq : forall x y z,
  x < y / z ->
  2 <= z ->
  word_to_nat (z * x) = (word_to_nat z * word_to_nat x)%nat.
Proof.
  intros [x X] [y Y] [z Z] LT LE; wunfold.
  rewrite wordModEq.

  assert (E := inj_mult (Zabs_nat z) (Zabs_nat x)).
  rewrite Z_to_nat_to_Z in E.
  rewrite Z_to_nat_to_Z in E.
  rewrite <- E.
  rewrite nat_to_Z_to_nat; auto.
  omega.
  omega.

  wunfold.
  constructor.
  auto with * .
  rewrite wordModEq in LT.
  assert (LT' : (z * x < z * (y / z))%Z).
  apply Zmult_lt_compat_l; omega.
  assert (LE' : (z * (y / z) <= y)%Z).
  apply Z_mult_div_ge; omega.
  omega.

  (* show x / y is a word value *)
  constructor.
  assert (y / z >= 0)%Z.
  apply Z_div_ge0; omega.
  omega.
  assert (y / z < y)%Z.
  apply Z_div_lt.
  omega.
  destruct (eq_Z_dec y 0) as [E | NE].
  subst y; replace (0 / z)%Z with 0%Z in LT.
  rewrite wordModEq in LT; wauto.
  reflexivity.
  omega.
  omega.
Qed.


(*
Theorem wordSubEqGe : forall a b,
  a <= b ->
  word_to_Z (b - a) = (word_to_Z b - word_to_Z a)%Z.
Proof.
  intros a b; generalize dependent a.
  apply wordInd1 with (P := fun b =>
    forall a, a <= b -> word_to_Z (b - a) =
    (word_to_Z b - word_to_Z a)%Z);
  unfold wordLte, wordLt, wordSub.

  (* base case *)
  replace (word_to_Z 0) with 0%Z.
  intros a LE.
  assert (E : word_to_Z a = 0%Z).
  assert (B := wordBounds a); omega.
  clear LE; rewrite E; clear E b.
  rewrite Z_to_word_to_Z; auto.
  simpl; wauto.
  unfold 0; rewrite nat_to_word_to_Z; wauto.

  (* inductive case *)
  clear b; intros b Ind LT a LE.
  destruct (eq_Z_dec (word_to_Z a) (word_to_Z b)).
  clear Ind LE; rewrite e; clear e a.
  rewrite Z_to_word_to_Z; auto.
  replace (word_to_Z b - word_to_Z b)%Z with 0%Z; wauto; omega.
  assert (LE' : (word_to_Z a <= word_to_Z (b - 1))%Z).
  apply lePred; unfold wordLt; omega.
  hyr Ind (Ind _ LE'); clear LE'.
  rewrite (wordSubEq LT) in Ind.
  replace (word_to_Z b - 1 - word_to_Z a)%Z with
    (word_to_Z b - word_to_Z a - 1)%Z in Ind; [ | omega].
  replace (word_to_Z (Z_to_binary (word_to_Z b - word_to_Z a - 1)))
    with (word_to_Z (Z_to_binary (word_to_Z b - word_to_Z a)) - 1)%Z in Ind; [omega| ].
  apply wpredEq; assert (B := wordBounds b); assert (A := wordBounds a); omega.
Qed.


Lemma wpredLtEq : forall w,
  w < w - 1 ->
  w = 0.
Proof.
  intros w LT; destruct (eq_Z_dec (word_to_Z w) (word_to_Z 0)).
  apply word_to_Z_inj; auto.
  replace (word_to_Z 0) with (0%Z) in n.
  assert (X := wordBounds w).
  elf; unfold wordLt in LT; rewrite wordSubEqGe in LT.
  unfold word1 in LT; rewrite nat_to_word_to_Z in LT; simpl in LT.
  omega.
  wauto.
  clear LT.
  unfold wordLte, word1; rewrite nat_to_word_to_Z; simpl; [omega | wauto].
  unfold 0; rewrite nat_to_word_to_Z; wauto.
Qed.

*)





Definition wordAlign (w:word) := w - w mod 4.

Lemma wordLeBound : forall x y,
  y <= x + y ->
  (word_to_Z x < wordBound - word_to_Z y)%Z.
Proof.
  intros [x xOk] [y yOk] LE; hyr LE (wordLeInj LE); simpl in LE; 
    unfold okWordZ in * .
  replace (word_to_Z (wordi xOk)) with x by auto.
  replace (word_to_Z (wordi yOk)) with y by auto.
  destruct (eq_Z_dec y 0) as [E | NE].

  omega.

  assert (D : (x + y < wordBound \/ 
    wordBound <= x + y < wordBound + wordBound - 1)%Z).
  omega.
  destruct D as [LT | [LE' LT]]; auto.

  omega.
  
  assert (E:(exists z, x + y = z + 1 * wordBound /\ 0 <= z < wordBound - 1)%Z).
  exists ((x + y) - wordBound)%Z.
  constructor; omega.
  destruct E as [z [E okz]].
  rewrite E in LE.
  rewrite Z_mod_plus in LE; wauto.
  rewrite wordModEq in LE.
  omega.
  unfold okWordZ; omega.
Qed.


Lemma wordAddEq : forall x y,
  y <= x + y ->
  word_to_Z (x + y) = (word_to_Z x + word_to_Z y)%Z.
Proof.
  intros [x xOk] [y yOk] LE; hyr LE (wordLeInj LE); wunfold.
  destruct (eq_Z_dec y 0) as [E | NE].

  subst; rewrite wordModEq; auto; wunfold ; omega.

  rewrite wordModEq; auto.
  wunfold.
  constructor.
  omega.
  assert (D : (x + y < wordBound \/ 
    wordBound <= x + y < wordBound + wordBound - 1)%Z).
  omega.
  destruct D as [LT | [LE' LT]]; auto; elf.
  
  assert (E:(exists z, x + y = z + 1 * wordBound /\ 0 <= z < wordBound - 1)%Z).
  exists ((x + y) - wordBound)%Z.
  constructor; omega.
  destruct E as [z [E okz]].
  rewrite E in LE.
  rewrite Z_mod_plus in LE; wauto.
  rewrite wordModEq in LE.
  omega.
  unfold okWordZ; omega.
Qed.


Lemma wordLeInj' : forall x y,
  (wval x <= wval y)%Z ->
  x <= y.
Proof.
  intros [x xOk] [y yOk] LE; hyr LE (wordLeInj LE); wunfold; auto.
Qed.


Lemma word_to_Z_natEq : forall x,
  Zabs_nat (word_to_Z x) = word_to_nat x.
Proof.
  intros [x xOk]; wunfold; auto.
Qed.


Lemma Zabs_natAddInj : forall x y,
  (0 <= x)%Z ->
  (0 <= y)%Z ->
  Zabs_nat (x + y)%Z = (Zabs_nat x + Zabs_nat y)%nat.
Proof.
  intros x y LEx LEy.
  destruct x.
  simpl; auto.
  destruct y; simpl.
  omega.
  rewrite nat_of_P_plus_morphism; auto.
  elf; compute in LEy; auto.
  elf; compute in LEx; auto.
Qed.


Lemma wordAddNatEq : forall y x,
  y <= x + y ->
  word_to_nat (x + y) = (word_to_nat x + word_to_nat y)%nat.
Proof.
  intros [y yOk] [x xOk] LT; assert (X := wordAddEq LT); clear LT.
  rewrite <- word_to_Z_natEq; rewrite X; clear X.
  wunfold; apply Zabs_natAddInj; omega.
Qed.


Lemma modOkWordZ : forall a b,
  okWordZ b ->
  (b > 0)%Z ->
  okWordZ (a mod b).
Proof.
  wunfold; intros a b LE pos.
  assert (0 <= a mod b < b)%Z.
  apply Z_mod_lt; omega.
  omega.
Qed.


Lemma modAddEq : forall x y,
  (y > 0)%Z ->
  (x mod y = (x + y) mod y)%Z.
Proof.
  intros x y GT.
  rewrite <- (Z_mod_plus x 1 _ GT).
  replace (1 * y)%Z with y; omega.
Qed.


Lemma wordModAdd4Eq : forall x,
  4 <= x + 4 ->
  (x + 4) mod 4 = x mod 4.
Proof.
  intros x LE; apply wordInj; destruct x as [x xOk].
  assert (E := wordAddEq LE).
  simpl in * ; rewrite E; clear LE E.
  wunfold; replace ((x + 4) mod 4)%Z with (x mod 4)%Z; auto.
  apply modAddEq; compute; auto.
Qed.


Lemma wordAlignAddEq : forall x,
  4 <= x + 4 ->
  wordAlign (x + 4) = wordAlign x + 4.
Proof.
  intros x LE.
  unfold wordAlign at 1; unfold wordAlign.
  replace ((x + 4) mod 4) with (x mod 4).
  ring.
  rewrite wordModAdd4Eq; auto.
Qed.


Lemma wordAlignLe : forall i,
  4 <= i + 4 ->
  4 <= wordAlign i + 4.
Proof.
  intros [i iOk] LE; apply wordLeInj'; hyr LE (wordLeInj LE); wunfold.
  assert (i4Ok : (0 <= i mod 4 < 4)%Z).
  apply Z_mod_lt; compute; auto.
  assert (LT : (4 < wordBound)%Z).
  compute; auto.
  rewrite (@wordModEq (i mod 4)).
  2:unfold okWordZ; omega.
  rewrite (@wordModEq (i - i mod 4)).

  clear LT.
  assert (D : (i < wordBound - 4 \/ wordBound - 4 <= i)%Z).
  omega.
  destruct D as [LT | GE].

  clear LE; rewrite wordModEq.
  destruct iOk as [iOk _]; clear LT.
  assert (i - i mod 4 >= 0)%Z.
  2:omega.
  assert (D : (i < 4 \/ i >= 4)%Z).
  omega.
  destruct D as [LE' | GT].
  assert (i = 0 \/ i = 1 \/ i = 2 \/ i = 3)%Z.
  omega.
  decompose [or] H; subst; compute; discriminate.
  omega.
  constructor; omega.
  
  elf.
  assert (i = wordBound - 4 \/ i = wordBound - 3 \/ i = wordBound - 2 \/
    i = wordBound - 1)%Z.
  omega.
  decompose [or] H; clear H; subst i; compute in LE; auto.

  constructor.
  2:omega.

  assert (D : (i < 4 \/ i >= 4)%Z).
  omega.
  destruct D as [LE' | GT].
  assert (i = 0 \/ i = 1 \/ i = 2 \/ i = 3)%Z.
  omega.
  decompose [or] H; subst; compute; discriminate.
  omega.

Qed.


Lemma word_to_natAlignEq : forall x,
  natWordAlign (word_to_nat x) = word_to_nat (wordAlign x).
Proof.
  induction x using wordInd4.
  compute; auto.
  compute; auto.
  compute; auto.
  compute; auto.

  assert (LE : 4 <= wordAlign x + 4).
  apply wordAlignLe; auto.
  rewrite wordAlignAddEq; auto.
  rewrite wordAddNatEq; auto.
  rewrite wordAddNatEq; auto.
  replace (word_to_nat x + word_to_nat 4)%nat with
    (word_to_nat 4 + word_to_nat x)%nat by auto with arith.
  simpl; rewrite IHx; auto.
  replace (word_to_nat 4) with 4%nat by auto; omega.
Qed.


Lemma wordLtInj' : forall (x y:word),
  (wval x < wval y)%Z ->
  x < y.
Proof.
  intros [x H] [y H'] LT; wunfold; auto.
Qed.


Lemma wordAlignOk : forall i,
  okWordZ i ->
  okWordZ (i - i mod 4).
Proof.
  unfold okWordZ; intros i iOk.
  assert (iOk' : (0 <= i mod 4 < 4)%Z).
  apply Z_mod_lt; compute; auto.
  assert (i = 0 \/ i = 1 \/ i = 2 \/ i = 3 \/ i >= 4)%Z.
  omega.
  decompose [or] H; subst; clear H.
  compute; auto.
  compute; constructor; auto; discriminate.
  compute; constructor; auto; discriminate.
  compute; constructor; auto; discriminate.
  omega.
Qed.


Lemma wordAlignLtInj : forall i,
  i < i + 4 ->
  wordAlign i < wordAlign (i + 4).
Proof.
  intros i LT.
  
  rewrite wordAlignAddEq.
  destruct i as [i iOk]; hyr LT (wordLtInj LT); apply wordLtInj'; wunfold.
  assert (iOk' : (0 <= i mod 4 < 4)%Z).
  apply Z_mod_lt; compute; auto.
  assert (i'Ok' : (0 <= (i + 4) mod 4 < 4)%Z).
  apply Z_mod_lt; compute; auto.
  assert (LT4 : (4 < wordBound)%Z).
  compute; auto.
  rewrite (@wordModEq (i mod 4)).
  2:constructor; omega.
  clear i'Ok'.
  rewrite (@wordModEq (i - i mod 4)).
  2:apply wordAlignOk; auto.
  rewrite wordModEq.
  omega.
  constructor.
  omega.
  assert (i + 4 < wordBound)%Z.
  2:omega.
  assert (i + 4 < wordBound \/ i = wordBound - 3 \/ i = wordBound - 2 \/
    i = wordBound - 1 \/ i = wordBound - 4)%Z.
  omega.
  decompose [or] H; clear H; auto; subst; elf; compute in LT; discriminate LT.

  destruct i as [i iOk]; apply wordLeInj'; hyr LT (wordLtInj LT); wunfold.
  assert (i + 4 < wordBound \/ i = wordBound - 3 \/ i = wordBound - 2 \/
    i = wordBound - 1 \/ i = wordBound - 4)%Z.
  omega.
  decompose [or] H; clear H; subst.
  rewrite wordModEq; unfold okWordZ; omega.
  compute in LT; discriminate LT.
  compute in LT; discriminate LT.
  compute in LT; discriminate LT.
  compute in LT; discriminate LT.
Qed.


Lemma wordAlignLe' : forall x,
  wordAlign x <= x.
Proof.
  intros [x xOk]; apply wordLeInj'; wunfold.
  assert (0 <= x mod 4 < 4)%Z.
  apply Z_mod_lt; compute; auto.
  rewrite (@wordModEq (x mod 4)).
  rewrite wordModEq.
  omega.
  apply wordAlignOk; auto.
  unfold okWordZ; assert (4 < wordBound)%Z.
  compute; auto.
  omega.
Qed.


Lemma wordAddLtI : forall x y,
  x <= y -> forall z,
  z <= y + z ->
  z <> 0 ->
  x < y + z.
Proof.
  intros [x xOk] [y yOk] LE [z zOk] LE' NE; apply wordLtInj'.
  hyr LE' (wordAddEq LE'); hyr LE (wordLeInj LE); hyr NE (wordNeInj NE); 
    wunfold; omega.
Qed.


Lemma wordAlignAlignEq : forall x,
  wordAlign (wordAlign x) = wordAlign x.
Proof.
  intros [x xOk]; apply wordInj; unfold wordAlign; wunfold.
  assert (0 <= x mod 4 < wordBound)%Z.
  assert (0 <= x mod 4 < 4)%Z.
  apply Z_mod_lt; wauto.
  assert (4 < wordBound)%Z.
  compute; auto.
  omega.
  rewrite (@wordModEq (x mod 4)).
  2:auto.
  rewrite (@wordModEq (x - x mod 4)).
  2:apply wordAlignOk; auto.
  rewrite (@wordModEq ((x - x mod 4) mod 4)).
  replace ((x - x mod 4) mod 4)%Z with 0%Z.
  rewrite wordModEq.
  omega.
  replace (x - x mod 4 - 0)%Z with (x - x mod 4)%Z by omega.
  apply wordAlignOk; auto.
  
  pattern x at 1; rewrite (Z_div_mod_eq x 4).
  2:omega.
  replace (4 * (x / 4) + x mod 4 - x mod 4)%Z with
    (0 + (x / 4) * 4)%Z by ring; rewrite Z_mod_plus; compute; auto.
  assert (0 <= (x - x mod 4) mod 4 < 4)%Z.
  apply Z_mod_lt; omega.
  assert (4 < wordBound)%Z.
  compute; auto.
  unfold okWordZ; omega.
Qed.


Lemma wordPlusLeNE : forall i,
  4 <= i + 4 ->
  0 <> wordAlign i + 4.
Proof.
  intros [i iOk] LE; apply wordNeInj'; hyr LE (wordLeBound LE); wunfold.
  assert (0 <= i mod 4 < 4)%Z.
  apply Z_mod_lt; compute; auto.
  rewrite (@wordModEq (i mod 4)).
  2:assert (4 < wordBound)%Z; [compute; auto | constructor; try omega].
  rewrite (@wordModEq (i - i mod 4)).
  2:apply wordAlignOk; auto.
  replace (i - i mod 4 + 4)%Z with ((i + 4) - i mod 4)%Z by ring.
  rewrite wordModEq.
  omega.
  unfold okWordZ.
  constructor.
  omega.
  cut (i < wordBound - 4)%Z.
  intro; omega.
  auto.
Qed.


Lemma anotherWordAlignLemma : forall i i',
  4 <= i + 4 ->
  wordAlign (i + 4) <= i' ->
  wordAlign i < i'.
Proof.
  intros [i iOk] [i' i'Ok] LE AL; hyr LE (wordLeBound LE); 
    hyr AL (wordLeInj AL); apply wordLtInj'; wunfold.
  assert (0 <= i mod 4 < 4)%Z.
  apply Z_mod_lt; omega.
  assert (4 < wordBound)%Z.
  compute; auto.
  assert (LE' : (i < wordBound - 4)%Z); [auto | clear LE; rename LE' into LE].
  generalize dependent AL.
  rewrite (@wordModEq (i + 4)).
  replace ((i + 4) mod 4)%Z with (i mod 4)%Z.
  rewrite (@wordModEq (i mod 4)).
  rewrite (@wordModEq (i - i mod 4)).
  rewrite wordModEq.
  intro; omega.
  2:apply wordAlignOk; auto.
  constructor; omega.
  constructor; omega.
  2:constructor; omega.
  replace (i + 4)%Z with (i + 1 * 4)%Z by omega.
  rewrite Z_mod_plus; omega.
Qed.


Lemma noOverflowLeToLt : forall x,
  4 <= x + 4 ->
  x < x + 4.
Proof.
  intros [x xOk] LE; hyr LE (wordLeBound LE); apply wordLtInj'; wunfold.
  rewrite wordModEq.
  omega.
  constructor; try omega.
  assert (x < wordBound - 4)%Z; [auto | omega].
Qed.


Lemma wordInj' : forall x y,
  x = y ->
  wval x = wval y.
Proof.
  intros; subst; auto.
Qed.


Lemma wordAlignPred : forall x y,
  4 <= y + 4 ->
  wordAlign x = x ->
  wordAlign x < wordAlign (y + 4) ->
  wordAlign x <> wordAlign y ->
  wordAlign x < wordAlign y.
Proof.
  intros [x xOk] [y yOk] LE X LT NE; hyr LT (wordLtInj LT); 
    hyr NE (wordNeInj NE); assert (E := wordAddEq LE); 
      assert (LE' := wordLeBound LE); clear LE; wunfold.
  unfold wordAlign in X.
  assert (E' : wval (wordi xOk - wordi xOk mod 4) = wval (wordi xOk)).
  rewrite X; auto.
  clear X; simpl in E'.
  generalize dependent E'.
  generalize dependent LT; generalize dependent NE; rewrite E.
  assert (0 <= x mod 4 < 4)%Z.
  apply Z_mod_lt; omega.
  assert (0 <= y mod 4 < 4)%Z.
  apply Z_mod_lt; omega.
  assert (4 < wordBound)%Z.
  compute; auto.
  rewrite (@wordModEq (x mod 4)).
  2:constructor; omega.
  rewrite (@wordModEq (y mod 4)).
  2:constructor; omega.
  rewrite wordModEq.
  2:apply wordAlignOk; auto.
  rewrite <- modAddEq; [ | omega].
  rewrite (@wordModEq (y mod 4)).
  2:constructor; omega.
  rewrite wordModEq.
  2:apply wordAlignOk; auto.
  rewrite wordModEq.

  destruct (alignProdEq x) as [k Ex].
  destruct (alignProdEq y) as [k' Ey].
  intros; omega.
  
  assert (X : (y < wordBound - 4)%Z); [auto | constructor; omega].

Qed.


Lemma Z_to_nat_to_word : forall x,
  (0 <= x)%Z ->
  nat_to_word (Zabs_nat x) = Z_to_word x.
Proof.
  intros x Pos; apply wordInj.
  simpl; rewrite Z_to_nat_to_Z; auto; omega.
Qed.


Lemma wordInj_sub : forall x y,
  Z_to_word (x - y) = Z_to_word x - Z_to_word y.
Proof.
  intros x y; apply wordInj.
  wunfold.
  pattern x at 1; rewrite (Z_div_mod_eq x wordBound). 2:wauto.
  pattern y at 1; rewrite (Z_div_mod_eq y wordBound). 2:wauto.
  replace (wordBound * (x / wordBound) + x mod wordBound -
     (wordBound * (y / wordBound) + y mod wordBound))%Z with
  ((x mod wordBound - y mod wordBound) + (x / wordBound - y / wordBound) * 
    wordBound)%Z by ring.
  rewrite Z_mod_plus; wauto.
Qed.


Lemma nat_to_word_inj : forall x y,
  okWordNat x -> okWordNat y ->
  nat_to_word x = nat_to_word y ->
  x = y.
Proof.
  intros x y okX okY E; hyr E (wordInj' E); simpl in E.
  rewrite wordModEq in E. 2:auto.
  rewrite wordModEq in E. 2:auto.
  omega.
Qed.


Lemma natWordAddEq : forall (x y:nat),
  ((x + y)%nat:word) = (x + y)%word.
Proof.
  induction x; intro y.

  simpl; replace (0 + y) with (y:word) by ring; auto.

  simpl.
  rewrite nat_to_wordSEq; rewrite (@nat_to_wordSEq x).
  rewrite IHx; ring.

Qed.


Lemma natWordSubEq : forall x y,
  x >= y ->
  ((x - y)%nat:word) = (x - y)%word.
Proof.
  clear. induction x; intros y GE.
  
  simpl; destruct y.
  replace (0 - 0)%word with 0 by ring; auto.
  elf; omega.
  
  rewrite nat_to_wordSEq.
  destruct y.
  simpl; rewrite nat_to_wordSEq; ring.
  simpl; rewrite IHx. 2:omega.
  rewrite (@nat_to_wordSEq y); ring.
Qed.


Lemma timesModEq : forall x,
  ((4 * x) mod wordBound = 4 * (x mod indexBound))%Z.
Proof.
  intros x.
  pattern x at 1.
  rewrite (Z_div_mod_eq x indexBound)%Z.
  2:compute; reflexivity.
  rewrite Zmult_plus_distr_r.
  replace (4 * (indexBound * (x / indexBound)))%Z with
    (wordBound * (x / indexBound))%Z.
  transitivity ((4 * (x mod indexBound) + (x / indexBound) * wordBound)
    mod wordBound)%Z.
  f_equal; ring.
  rewrite Z_mod_plus.
  rewrite wordModEq.
  reflexivity.
  unfold okWordZ.
  assert (0 <= x mod indexBound < indexBound)%Z as LE'.
  apply Z_mod_lt; vm_compute; auto.
  destruct LE' as [xPos LT].
  constructor.
  omega.
  clear xPos.
  generalize dependent (x mod indexBound)%Z; clear; intros.
  2:wauto.
  replace indexBound%Z with 1073741824%Z in LT by reflexivity.
  replace wordBound with 4294967296%Z by reflexivity.
  omega.

  generalize dependent (x / indexBound)%Z; clear; intros.
  replace (4 * (indexBound * z))%Z with (4 * indexBound * z)%Z by ring.
  reflexivity.

Qed.


Lemma nat_to_wordAlign : forall k,
  nat_to_word (4 * k) =
  wordAlign (nat_to_word (4 * k)).
Proof.
  intros k.
  unfold wordAlign.
  replace (nat_to_word (4 * k) mod nat_to_word 4) with 0.
  replace ((4 * k)%nat - 0) with ((4 * k)%nat:word) by ring; auto.
  apply wordInj; simpl.
  rewrite zeroModEq; rewrite fourModEq.
  replace (k + (k + (k + (k + 0))))%nat with (k + k + k + k)%nat by omega.
  rewrite inj_plus; rewrite inj_plus; rewrite inj_plus.
  replace (Z_of_nat k + Z_of_nat k + Z_of_nat k + Z_of_nat k)%Z with
    (4 * Z_of_nat k)%Z by omega.
  generalize (Z_of_nat k); clear k; intro x.
  replace (((4 * x) mod wordBound) mod 4)%Z with 0%Z.
  rewrite zeroModEq; auto.
  symmetry; apply Z_div_exact_1.
  omega.
  rewrite timesModEq.
  replace (4 * (x mod indexBound) / 4)%Z with
    (x mod indexBound)%Z.
  auto.
  replace (4 * (x mod indexBound))%Z with
    (0 + (x mod indexBound) * 4)%Z by ring.
  rewrite Z_div_plus.
  auto.
  vm_compute; reflexivity.
Qed.


Lemma nat_to_word_to_natAlign : forall k,
  word_to_nat (nat_to_word (4 * k)) =
  natWordAlign (word_to_nat (nat_to_word (4 * k))).
Proof.
  intros k.
  rewrite word_to_natAlignEq; rewrite <- nat_to_wordAlign; auto.
Qed.


Lemma nat_to_word_to_natLE : forall x,
  (word_to_nat (nat_to_word x) <= x)%nat.
Proof.
  intros x; unfold nat_to_word, word_to_nat; simpl.
  assert (Z_of_nat x mod wordBound <= Z_of_nat x)%Z as LE by (apply modSmaller; wauto).
  assert (0 <= Z_of_nat x mod wordBound < wordBound)%Z as LE' by (apply Z_mod_lt; wauto).
  generalize dependent (Z_of_nat x mod wordBound)%Z; clear; intros.
  destruct LE' as [zPos _].
  rewrite <- nat_to_Z_to_nat.
  apply Zabs_natLeInj; auto.
Qed.


Lemma okWordNatLtTrans : forall x y,
  (x <= y)%nat ->
  okWordNat y ->
  okWordNat x.
Proof.
  unfold okWordNat; intros x y LE Ok; omega.
Qed.


Lemma wordAddInj : forall a b c,
  a + b = a + c ->
  b = c.
Proof.
  clear; intros a b c E.
  assert (a + b - a = a + c - a) as E'.
  rewrite E; auto.
  replace (a + b - a) with b in E' by ring.
  clear E; subst; ring.
Qed.


Lemma wordIndexMultInj : forall x y,
  x < Z_to_word indexBound ->
  y < Z_to_word indexBound ->
  4 * x = 4 * y ->
  x = y.
Proof.
  clear; intros [x xOk] [y yOk] LTx LTy E.
  unfold okWordZ in * .
  apply wordLtInj in LTx.
  apply wordLtInj in LTy.
  apply wordInj' in E; apply wordInj.
  simpl in LTx, LTy |- * .
  rewrite wordModEq in LTx.
  2:vm_compute; constructor; auto; discriminate.
  rewrite wordModEq in LTy.
  2:vm_compute; constructor; auto; discriminate.
  replace (wval (4 * wordi xOk)) with ((4 * x) mod wordBound)%Z in E by auto.
  replace (wval (4 * wordi yOk)) with ((4 * y) mod wordBound)%Z in E by auto.
  rewrite timesModEq in E; rewrite timesModEq in E.
  rewrite ltModEq in E; [ | omega | omega].
  rewrite ltModEq in E; omega.
Qed.


Lemma wordIndexMultLeInj : forall x y,
  (x < windexBound)%word ->
  (y < windexBound)%word ->
  (4 * x <= 4 * y)%word ->
  (x <= y)%word.
Proof.
  clear; intros [x xOk] [y yOk] LTx LTy LE.
  unfold okWordZ in * .
  apply wordLtInj in LTx. apply wordLtInj in LTy.
  apply wordLeInj in LE; apply wordLeInj'.
  simpl in LTx, LTy |- * .
  rewrite wordModEq in LTx. 2:auto with Word.
  rewrite wordModEq in LTy. 2:auto with Word.
  replace (wval (4 * wordi xOk)) with ((4 * x) mod wordBound)%Z in LE by auto.
  replace (wval (4 * wordi yOk)) with ((4 * y) mod wordBound)%Z in LE by auto.
  rewrite timesModEq in LE; rewrite timesModEq in LE.
  rewrite ltModEq in LE; [ | omega | omega].
  rewrite ltModEq in LE; omega.
Qed.


Lemma arrayLenBoundNoOverflow : forall x,
  (Z_of_nat (word_to_nat x + 1) <= indexBound)%Z ->
  0 <> x + 1.
Proof.
  clear; intros x LE E.
  replace (word_to_nat x + 1)%nat with (S (word_to_nat x)) in LE by omega.
  rewrite inj_S in LE; unfold Zsucc in LE.
  destruct x as [x xOk]; apply wordInj' in E; wunfold.
  rewrite Z_to_nat_to_Z in LE.
  rewrite wordModEq in E.
  omega.
  constructor.
  omega.
  assert (indexBound < wordBound)%Z by (vm_compute; auto); omega.
  omega.
Qed.


Lemma bool_to_word_to_bool : forall b,
  word_to_bool (bool_to_word b) = b.
Proof.
  unfold word_to_bool; destruct b; simpl.
  destruct (eq_word_dec 1 0) as [E | NE]; [discriminate | auto].
  destruct (eq_word_dec 0 0) as [E | NE]; [ | elf]; auto.
Qed.


Lemma indexBoundOk : okWordZ indexBound.
Proof.
  compute; intuition; discriminate.
Qed.
Hint Resolve indexBoundOk : Word.


Lemma word_to_nat_to_Z : forall x,
  Z_of_nat (word_to_nat x) = wval x.
Proof.
  intros [x xOk]; unfold okWordZ in xOk; simpl; unfold word_to_nat;
    rewrite Z_to_nat_to_Z; auto.
  simpl; omega.
Qed.




Ltac uomega := unfold okWordZ, wordBound, indexBound in *; omega.


(* womega attempts to turn a goal involving words into a goal involving Z,
   and does some additional simplifications *)

Lemma nat_to_wordWvalEq : forall k,
  wval (nat_to_word k) = ((Z_of_nat k) mod wordBound)%Z.
Proof.
  reflexivity.
Qed.

Lemma wordiWvalEq : forall w (H:okWordZ w),
  wval (wordi H) = w.
Proof.
  reflexivity.
Qed.

Hint Rewrite nat_to_wordWvalEq wordiWvalEq : WordZ.
Hint Rewrite zeroModEq oneModEq twoModEq threeModEq fourModEq : WordZ.
Hint Rewrite inj_plus inj_S inj_0 : WordNatZ.
Hint Rewrite Z_to_nat_to_Z using omega : WordNatZ.


Ltac wordUnfold :=
  unfold wordAlign, wordNeg, wordSub, wordAdd, wordMult, wordDiv, wordMod, 
    wordOp, 
    wordLt, wordLte, wordP,
    okWordZ, okWordNat, word_to_nat,
    wordBound, indexBound in * .

Lemma wordModEq' : forall a,
  (0 <= a < 4294967296)%Z ->
  (a mod 4294967296)%Z = a.
Proof.
  intros a X; apply wordModEq; auto.
Qed.

 
Lemma wordiWvalBoundModEq : forall x,
  (wval (wordi (boundModOk x)) = x mod wordBound)%Z.
Proof.
  reflexivity.
Qed.

Lemma zeroModEq' : (Z_of_nat 0 mod wordBound = 0)%Z.
Proof.
  simpl; rewrite zeroModEq; reflexivity.
Qed.

Lemma oneModEq' : (Z_of_nat 1 mod wordBound = 1)%Z.
Proof.
  simpl; rewrite oneModEq; reflexivity.
Qed.

Lemma twoModEq' : (Z_of_nat 2 mod wordBound = 2)%Z.
Proof.
  simpl; rewrite twoModEq; reflexivity.
Qed.

Lemma threeModEq' : (Z_of_nat 3 mod wordBound = 3)%Z.
Proof.
  simpl; rewrite threeModEq; reflexivity.
Qed.

Lemma fourModEq' : (Z_of_nat 4 mod wordBound = 4)%Z.
Proof.
  simpl; rewrite fourModEq; reflexivity.
Qed.

Lemma indexBoundModEq :
  (indexBound mod wordBound = indexBound)%Z.
Proof.
  rewrite wordModEq; wauto.
Qed.


Hint Rewrite zeroModEq' oneModEq' twoModEq' threeModEq' fourModEq' 
  wordiWvalBoundModEq timesModEq indexBoundModEq : WordZ.


Lemma multModEq : forall y x,
  (y > 0)%Z ->
  ((y * x) mod y = 0)%Z.
Proof.
  intros y x GT.
  replace (y * x)%Z with (0 + x * y)%Z by ring.
  rewrite Z_mod_plus; auto.
Qed.

Lemma fourMultModEq : forall x,
  ((4 * x) mod 4 = 0)%Z.
Proof.
  intros x; apply multModEq; omega.
Qed.
Hint Rewrite fourMultModEq : WordZ.

Lemma minusZeroEq : forall x,
  (x - 0 = x)%Z.
Proof.
  intros; omega.
Qed.
Hint Rewrite minusZeroEq : WordZ.

Hint Rewrite modModEqEq using (wauto; omega) : WordZ.

Lemma wordIndexModEq' : forall a,
  (0 <= a < 1073741824)%Z ->
  (a mod 1073741824)%Z = a.
Proof.
  intros a H; rewrite ltModEq; omega.
Qed.



Ltac natLeSolve :=
  apply inj_le_rev; autorewrite with WordNatZ; omega.

Ltac natLtSolve :=
  apply inj_lt_rev; autorewrite with WordNatZ; omega.

Ltac natEqSolve :=
  apply inj_eq_rev; autorewrite with WordNatZ; omega.


Hint Rewrite inj_minus1 using natLeSolve : WordNatZ.

Hint Rewrite wordiWvalEq : WordNatZ.



Ltac womega :=
  first [apply wordNeInj' | apply wordInj | idtac ];
    repeat match goal with 
             | H : (_ <> _) |- _ => apply wordNeInj in H
             | H : (_ = _) |- _ => apply wordInj' in H
             | x : word |- _ => let xOk := fresh "wOk" in destruct x as [x xOk]
           end;
    unfold wordAlign, wordNeg, wordSub, wordAdd, wordMult, wordDiv, 
      windexBound, Z_to_word,
      wordMod, wordOp, wordLt, wordLte, wordP in *;
    autorewrite with WordZ in * ;
    unfold okWordZ, okWordNat, word_to_nat, wordBound, indexBound in *;
    autorewrite with WordNatZ in * ;
    repeat match goal with 
             | |- context [(?x mod 4294967296)%Z] =>
               match x with
                 | context [(_ mod _)%Z] => fail 1
                 | _ => rewrite (@wordModEq' x); [ | womega; fail]
               end
             | |- context [(?x mod 1073741824)%Z] =>
               match x with
                 | context [(_ mod _)%Z] => fail 1
                 | _ => rewrite (@wordIndexModEq' x); [ | womega; fail]
               end
           end;
    repeat match goal with 
             | H : context [(?x mod 4294967296)%Z] |- _ =>
               match x with
                 | context [(_ mod _)%Z] => fail 1
                 | _ => rewrite (@wordModEq' x) in H; [ | uomega]
               end
             | H : context [(?x mod 1073741824)%Z] |- _ =>
               match x with
                 | context [(_ mod _)%Z] => fail 1
                 | _ => rewrite (@wordIndexModEq' x) in H; [ | uomega]
               end
           end;
    try rewrite inj_plus in *; (* kind of hacky *)
    first [natLeSolve | natLtSolve | natEqSolve | try omega].



