(******************************************************************

   Separation logic toolkit

   Andrew McCreight
   HASP Project
   Portland State University

   http://hasp.cs.pdx.edu/

   Copyright 2009 Portland State University

   This file is part of the separation logic toolkit.

   The separation logic toolkit is free software: you can redistribute
   it and/or modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation, either
   version 3 of the License, or (at your option) any later version.

   The separation logic toolkit is distributed in the hope that it
   will be useful, but WITHOUT ANY WARRANTY; without even the implied
   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with the separation logic toolkit.  If not, see
   <http://www.gnu.org/licenses/>.

*******************************************************************)


Require Import ZArith.
Require Export MyTactics.


Ltac elf := elimtype False.

Open Local Scope Z_scope.


Lemma zlt_dec : forall m n:Z,
  {m < n} + {~m < n}.
Proof.
  intros m n.
  assert (X : match m ?= n with
		| Eq => m = n
		| Lt => m < n
		| Gt => m > n
	      end).
  apply (Zcompare_elim (m = n) (m < n) (m > n) m n); auto.
  destruct (m ?= n).
  subst; right; omega.
  auto.
  right; omega.
Qed.


Lemma two_powerEq : forall n,
  two_p (Z_of_nat n) = two_power_nat n.
Proof.
  induction n; auto.
  rewrite inj_S; rewrite two_p_S.
  rewrite IHn; auto.
  omega.
Qed.


Lemma two_power_natPos : forall n,
  two_power_nat n > 0.
Proof.
  intros n.
  rewrite <- two_powerEq.
  assert (X := two_p_gt_ZERO (Z_of_nat n)); omega.
Qed.


Lemma eq_Z_dec : forall m n:Z,
  {m = n} + {m <> n}.
Proof.
  intros m n.
  assert (X : match m ?= n with
		| Eq => m = n
		| Lt => (m < n)
		| Gt => (m > n)
	      end).
  apply (Zcompare_elim (m = n) (m < n) (m > n) m n); auto.
  destruct (m ?= n); auto; right; omega.
Qed.


(* conversion between nat and Z *)

Lemma nat_to_Z_to_nat : forall k,
  Zabs_nat (Z_of_nat k) = k.
Proof.
  destruct k; simpl; auto; apply nat_of_P_o_P_of_succ_nat_eq_succ.
Qed.

Lemma Z_to_nat_to_Z : forall k,
  k >= 0 ->
  Z_of_nat (Zabs_nat k) = k.
Proof.
  intros k GE.
  induction k; simpl; auto.
  symmetry; apply Zpos_eq_Z_of_nat_o_nat_of_P.
  elf; destruct GE; auto.
Qed.


Lemma ltDivEq : forall a b,
  0 <= a < b ->
  b > 0 ->
  a / b = 0.
Proof.
  intros a b [aPos abLt] bPos.
  assert (D : a / b = 0 \/ a / b < 0 \/ a / b > 0).
  omega.
  destruct D as [E | [LT | GT]]; auto.
  assert (Y : a >= 0); [omega | ].
  assert (X := Z_div_ge0 _ _ bPos Y); clear Y; omega.
  elf.
  assert (LE := Z_mult_div_ge a b bPos).
  assert (b  * (a / b) >= b * 1).
  assert (a / b >= 1).
  omega.
  apply Zmult_ge_compat_l; auto; omega.
  omega.
Qed.


Lemma ltModEq : forall a b,
  0 <= a < b ->
  b > 0 ->
  a mod b = a.
Proof.
  intros a b aOk bOk.
  transitivity (b * (a / b) + a mod b).
  rewrite ltDivEq; auto; omega.
  rewrite <- Z_div_mod_eq; auto.
Qed.


Lemma predModEq : forall x y,
  x < y -> forall z,
  0 <= x < z ->
  0 <= y < z ->
  ((y - 1) mod z = y - 1)%Z.
Proof.
  intros x y LT z Ok1 Ok2.
  apply ltModEq; omega.
Qed.


Lemma boundPre : forall n z,
  0 <= Z_of_nat (S n) < z ->
  0 <= Z_of_nat n < z.
Proof.
  intros n z LT.
  replace (Z_of_nat (S n)) with (Z_of_nat n + 1)%Z in LT.
  omega.
  clear LT; unfold Z_of_nat; destruct n; auto; simpl.
  rewrite Pplus_one_succ_r; auto.
Qed.


Lemma negModEq : forall a b,
  (a < 0)%Z ->
  (b > 0)%Z ->
  (a mod b <> 0)%Z ->
  (a mod b = b - (-a) mod b)%Z.
Proof.
  intros a b LT GT NE.
  destruct a; try (inversion LT).
  destruct b; try (inversion GT).
  unfold Zmod, Zdiv_eucl in * .
  replace (-Zneg p)%Z with (Zpos p) by auto.
  destruct (Zdiv_eucl_POS p (Zpos p0)).
  destruct z0; auto; elf; auto.
Qed.


Lemma modModEqEq : forall a b,
  (b > 0)%Z ->
  ((a mod b) mod b = a mod b)%Z.
Proof.
  intros a b Pos.
  pattern a at 2.
  rewrite (Z_div_mod_eq a _ Pos).
  rewrite Zplus_comm.
  rewrite Zmult_comm.
  rewrite Z_mod_plus; auto.
Qed.


Lemma modPlusEq : forall a b c,
  c > 0 ->
  ((a mod c + b mod c) mod c)%Z =
  ((a + b) mod c)%Z.
Proof.
  intros a b c cPos.
  pattern a at 2; rewrite (Z_div_mod_eq a c); auto.
  pattern b at 2; rewrite (Z_div_mod_eq b c); auto.
  replace (c * (a / c) + a mod c + (c * (b / c) + b mod c))%Z with
    ((a mod c + b mod c) + ((a / c + b / c) * c))%Z by ring.
  rewrite Z_mod_plus; auto.
Qed.


Lemma modMinusEq : forall a b c,
  c > 0 ->
  ((a mod c - b mod c) mod c)%Z =
  ((a - b) mod c)%Z.
Proof.
  intros a b c cPos.
  pattern a at 2; rewrite (Z_div_mod_eq a c); auto.
  pattern b at 2; rewrite (Z_div_mod_eq b c); auto.
  replace (c * (a / c) + a mod c - (c * (b / c) + b mod c))%Z with
    ((a mod c - b mod c) + ((a / c - b / c) * c))%Z by ring.
  rewrite Z_mod_plus; auto.
Qed.


Lemma modMultEq : forall a b c,
  (c > 0)%Z ->
  ((a mod c * (b mod c)) mod c)%Z =
  ((a * b) mod c)%Z.
Proof.
  intros a b c cPos.
  pattern a at 2; rewrite (Z_div_mod_eq a c); auto.
  pattern b at 2; rewrite (Z_div_mod_eq b c); auto.
  replace ((c * (a / c) + a mod c) * (c * (b / c) + b mod c))%Z with
    ((a mod c * (b mod c)) + ((c * (a / c) * (b / c)) + 
      (b mod c) * (a / c) + (a mod c) * (b / c)) * c)%Z by ring.
  rewrite Z_mod_plus; auto.
Qed.


Lemma plusModDistrib : forall a b c,
  c > 0 ->
  ((a mod c + b) mod c = (a + b) mod c)%Z.
Proof.
  intros a b c cPos.
  pattern b at 1.
  rewrite (Z_div_mod_eq b c); auto.
  replace (a mod c + (c * (b / c) + b mod c))%Z
    with ((a mod c + b mod c) + (b / c) * c)%Z by ring.
  rewrite Z_mod_plus; auto; apply modPlusEq; auto.
Qed.


Lemma minusModDistrib : forall a b c,
  c > 0 ->
  ((a mod c - b) mod c = (a - b) mod c)%Z.
Proof.
  intros a b c cPos.
  pattern b at 1.
  rewrite (Z_div_mod_eq b c); auto.
  replace (a mod c - (c * (b / c) + b mod c))%Z
    with ((a mod c - b mod c) + (-1 * (b / c)) * c)%Z by ring.
  rewrite Z_mod_plus; auto; apply modMinusEq; auto.
Qed.


Lemma multModDistrib : forall a b c,
  (c > 0)%Z ->
  ((a mod c * b) mod c = (a * b) mod c)%Z.
Proof.
  intros a b c cPos.
  pattern b at 1.
  rewrite (Z_div_mod_eq b c); auto.
  replace (a mod c * (c * (b / c) + b mod c))%Z
    with ((a mod c * (b mod c)) + (a mod c * (b / c)) * c)%Z by ring.
  rewrite Z_mod_plus; auto; apply modMultEq; auto.
Qed.



Lemma alignProdEq : forall x,
  (exists k, x - x mod 4 = 4 * k)%Z.
Proof.
  intros x; exists (x / 4)%Z.
  pattern x at 1.
  rewrite (Z_div_mod_eq x 4).
  ring.
  compute; auto.
Qed.


Lemma modSmaller : forall x y,
  x >= 0 ->
  y > 0 ->
  x mod y <= x.
Proof.
  intros x y xPos yPos.
  assert (x < y \/ x >= y) as D by omega.
  destruct D.
  rewrite ltModEq; omega.
  assert (0 <= x mod y < y) as LT by (apply Z_mod_lt; auto).
  omega.
Qed.


(* redundant with inj_le_rev? *)
Lemma Zabs_natLeInj : forall x y,
  0 <= x <= y ->
  (Zabs_nat x <= Zabs_nat y)%nat.
Proof.
  intros x y LT.
  assert (Zabs_nat x <= Zabs_nat y \/ Zabs_nat x > Zabs_nat y)%nat as D by omega.
  destruct D; auto.
  apply inj_gt in H.
  rewrite Z_to_nat_to_Z in H. 2:omega.
  rewrite Z_to_nat_to_Z in H; omega.
Qed.


(* redundant with inj_le_rev? *)
Lemma Zabs_natLtInj : forall x y,
  0 <= x < y ->
  (Zabs_nat x < Zabs_nat y)%nat.
Proof.
  intros x y LT.
  assert (Zabs_nat x < Zabs_nat y \/ Zabs_nat x >= Zabs_nat y)%nat as D by omega.
  destruct D; auto.
  apply inj_ge in H.
  rewrite Z_to_nat_to_Z in H. 2:omega.
  rewrite Z_to_nat_to_Z in H; omega.
Qed.



(* the Coq library is a bit incomplete for Zabs_nat *)
Lemma posZabsSucc : forall x,
  (0 <= x)%Z ->
  S (Zabs_nat x) = Zabs_nat (Zsucc x).
Proof.
  clear.
  induction x.
  reflexivity.
  intros _; simpl.
  rewrite <- nat_of_P_succ_morphism; f_equal.
  rewrite Pplus_one_succ_r; auto.
  intros GE; elf; compute in GE; auto.
Qed.


Lemma inj_0 : Z_of_nat 0 = 0%Z.
Proof.
  reflexivity.
Qed.